#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the S3 Artifact Store."""

from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

import boto3
import s3fs
from botocore.exceptions import ClientError
from fsspec.asyn import FSTimeoutError, sync, sync_wrapper

from zenml.artifact_stores import BaseArtifactStore
from zenml.integrations.s3.flavors.s3_artifact_store_flavor import (
    S3ArtifactStoreConfig,
)
from zenml.integrations.s3.utils import split_s3_path
from zenml.io.fileio import convert_to_str
from zenml.logger import get_logger
from zenml.secret.schemas import AWSSecretSchema
from zenml.stack.authentication_mixin import AuthenticationMixin

logger = get_logger(__name__)

PathType = Union[bytes, str]


def _normalize_s3_path(path: str) -> str:
    """Normalizes an S3 path by removing the protocol prefix if present.

    Args:
        path: The path to normalize.

    Returns:
        The normalized path.
    """
    if path.startswith("s3://"):
        return path[5:]
    return path


class ZenMLS3Filesystem(s3fs.S3FileSystem):  # type: ignore[misc]
    """Modified s3fs.S3FileSystem to disable caching.

    The original s3fs.S3FileSystem caches all class instances based on the
    constructor input arguments and it never releases them. This is problematic
    in the context of the ZenML server, because the server is a long-running
    process that instantiates many S3 filesystems with different credentials,
    especially when the credentials are generated by service connectors.

    The caching behavior of s3fs causes the server to slowly consume more and
    more memory over time until it crashes. This class disables the caching
    behavior of s3fs by setting the `cachable` attribute to `False`.

    In addition to disabling instance caching, this class also provides a
    correct cleanup implementation by overriding the `close_session` method
    the S3 aiobotocore client. The original one provided by s3fs was causing
    memory leaks by creating a new event loop in the destructor instead of
    using the existing one.

    A `close` method is also provided to allow for synchronous on-demand cleanup
    of the S3 client.
    """

    cachable = False

    @staticmethod
    async def _safe_aexit_s3_creator(s3_creator: Any) -> None:
        """Exit an aiobotocore client context manager, tolerating double-close.

        The s3fs finalizer and ZenML's explicit cleanup can both call
        ``__aexit__`` on the same creator object. The second call raises
        ``AssertionError("Session was never entered")`` in aiobotocore
        because ``http_session._sessions`` is already ``None``. This
        wrapper treats that specific assertion as a harmless no-op.

        Args:
            s3_creator: The aiobotocore async context manager to exit.

        Raises:
            AssertionError: If the assertion is unrelated to double-close.
        """
        try:
            await s3_creator.__aexit__(None, None, None)
        except AssertionError as exc:
            _MSG = "Session was never entered"
            if not (exc.args and _MSG in str(exc.args[0])):
                raise
            logger.debug(
                "Suppressed double-close of aiobotocore session: %s", exc
            )

    async def _close(self) -> None:
        """Close the S3 client."""
        if self._s3creator is not None:  # type: ignore[has-type]
            await self._safe_aexit_s3_creator(self._s3creator)  # type: ignore[has-type]
            self._s3creator = None
            self._s3 = None

    close = sync_wrapper(_close)

    @staticmethod
    def close_session(loop: Any, s3: Any) -> None:
        """Close the S3 client session.

        Args:
            loop: The event loop to use for closing the session.
            s3: The S3 client to close.
        """
        # IMPORTANT: This method is a copy of the original close_session method
        # from s3fs.S3FileSystem. The only difference is that it uses the
        # provided event loop instead of creating a new one, and routes
        # through _safe_aexit_s3_creator to tolerate double-close.
        if loop is not None and loop.is_running():
            try:
                # NOTE: this is the line in the original method that causes
                # the memory leak
                # loop = asyncio.get_event_loop()
                loop.create_task(ZenMLS3Filesystem._safe_aexit_s3_creator(s3))
                return
            except RuntimeError:
                pass
            try:
                sync(
                    loop,
                    ZenMLS3Filesystem._safe_aexit_s3_creator,
                    s3,
                    timeout=0.1,
                )
                return
            except FSTimeoutError:
                pass
        try:
            # close the actual socket
            s3._client._endpoint.http_session._connector._close()
        except AttributeError:
            # but during shutdown, it may have gone
            pass


class S3ArtifactStore(BaseArtifactStore, AuthenticationMixin):
    """Artifact Store for S3 based artifacts."""

    _filesystem: Optional[ZenMLS3Filesystem] = None

    is_versioned: bool = False

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initializes the artifact store.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._boto3_bucket_holder = None

        # determine bucket versioning status
        versioning = self._boto3_bucket.Versioning()
        with self._shield_lack_of_versioning_permissions(
            "s3:GetBucketVersioning"
        ):
            if versioning.status == "Enabled":
                self.is_versioned = True
                logger.warning(
                    f"The artifact store bucket `{self.config.bucket}` is versioned, "
                    "this may slow down logging process significantly."
                )

    @property
    def config(self) -> S3ArtifactStoreConfig:
        """Get the config of this artifact store.

        Returns:
            The config of this artifact store.
        """
        return cast(S3ArtifactStoreConfig, self._config)

    def get_credentials(
        self,
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        """Gets authentication credentials.

        If an authentication secret is configured, the secret values are
        returned. Otherwise, we fall back to the plain text component
        attributes.

        Returns:
            Tuple (key, secret, token, region) of credentials used to
            authenticate with the S3 filesystem.

        Raises:
            RuntimeError: If the AWS connector behaves unexpectedly.
        """
        connector = self.get_connector()
        if connector:
            from botocore.client import BaseClient

            client = connector.connect()
            if not isinstance(client, BaseClient):
                raise RuntimeError(
                    f"Expected a botocore.client.BaseClient while trying to "
                    f"use the linked connector, but got {type(client)}."
                )
            credentials = client.credentials
            return (
                credentials.access_key,
                credentials.secret_key,
                credentials.token,
                client.meta.region_name,
            )

        secret = self.get_typed_authentication_secret(
            expected_schema_type=AWSSecretSchema
        )
        if secret:
            return (
                secret.aws_access_key_id,
                secret.aws_secret_access_key,
                secret.aws_session_token,
                None,
            )
        else:
            return self.config.key, self.config.secret, self.config.token, None

    @property
    def filesystem(self) -> ZenMLS3Filesystem:
        """The s3 filesystem to access this artifact store.

        Returns:
            The s3 filesystem.
        """
        # Refresh the credentials also if the connector has expired
        if self._filesystem and not self.connector_has_expired():
            return self._filesystem

        key, secret, token, region = self.get_credentials()

        # Use the region from the connector if available, otherwise some
        # remote workloads (e.g. Sagemaker) might not work correctly because
        # they look for the bucket in the wrong region
        client_kwargs = {}
        if region:
            client_kwargs["region_name"] = region
        if self.config.client_kwargs:
            client_kwargs.update(self.config.client_kwargs)

        self._filesystem = ZenMLS3Filesystem(
            key=key,
            secret=secret,
            token=token,
            client_kwargs=client_kwargs,
            config_kwargs=self.config.config_kwargs,
            s3_additional_kwargs=self.config.s3_additional_kwargs,
        )
        return self._filesystem

    def cleanup(self) -> None:
        """Close the filesystem."""
        if self._filesystem:
            self._filesystem.close()

    def open(self, path: PathType, mode: str = "r") -> Any:
        """Open a file at the given path.

        Args:
            path: Path of the file to open.
            mode: Mode in which to open the file. Currently, only
                'rb' and 'wb' to read and write binary files are supported.

        Returns:
            A file-like object.
        """
        return self.filesystem.open(path=path, mode=mode)

    def copyfile(
        self, src: PathType, dst: PathType, overwrite: bool = False
    ) -> None:
        """Copy a file.

        Args:
            src: The path to copy from.
            dst: The path to copy to.
            overwrite: If a file already exists at the destination, this
                method will overwrite it if overwrite=`True` and
                raise a FileExistsError otherwise.

        Raises:
            FileExistsError: If a file already exists at the destination
                and overwrite is not set to `True`.
        """
        if not overwrite and self.filesystem.exists(dst):
            raise FileExistsError(
                f"Unable to copy to destination '{convert_to_str(dst)}', "
                f"file already exists. Set `overwrite=True` to copy anyway."
            )

        # TODO [ENG-151]: Check if it works with overwrite=True or if we need to
        #  manually remove it first
        self.filesystem.copy(path1=src, path2=dst)

    def exists(self, path: PathType) -> bool:
        """Check whether a path exists.

        Args:
            path: The path to check.

        Returns:
            True if the path exists, False otherwise.
        """
        return self.filesystem.exists(path=path)  # type: ignore[no-any-return]

    def glob(self, pattern: PathType) -> List[PathType]:
        """Return all paths that match the given glob pattern.

        The glob pattern may include:
        - '*' to match any number of characters
        - '?' to match a single character
        - '[...]' to match one of the characters inside the brackets
        - '**' as the full name of a path component to match to search
            in subdirectories of any depth (e.g. '/some_dir/**/some_file)

        Args:
            pattern: The glob pattern to match, see details above.

        Returns:
            A list of paths that match the given glob pattern.
        """
        return [f"s3://{path}" for path in self.filesystem.glob(path=pattern)]

    def isdir(self, path: PathType) -> bool:
        """Check whether a path is a directory.

        Args:
            path: The path to check.

        Returns:
            True if the path is a directory, False otherwise.
        """
        return self.filesystem.isdir(path=path)  # type: ignore[no-any-return]

    def listdir(self, path: PathType) -> List[PathType]:
        """Return a list of files in a directory.

        Args:
            path: The path to list.

        Returns:
            A list of paths that are files in the given directory.
        """
        # remove s3 prefix if given, so we can remove the directory later as
        # this method is expected to only return filenames
        path = _normalize_s3_path(convert_to_str(path))

        def _extract_basename(file_dict: Dict[str, Any]) -> str:
            """Extracts the basename from a file info dict returned by the S3 filesystem.

            Args:
                file_dict: A file info dict returned by the S3 filesystem.

            Returns:
                The basename of the file.
            """
            file_path = cast(str, file_dict["Key"])
            base_name = file_path[len(path) :]
            return base_name.lstrip("/")

        return [
            _extract_basename(dict_)
            for dict_ in self.filesystem.listdir(path=path)
            # s3fs.listdir also returns the root directory, so we filter
            # it out here
            if _extract_basename(dict_)
        ]

    def makedirs(self, path: PathType) -> None:
        """Create a directory at the given path.

        If needed also create missing parent directories.

        Args:
            path: The path to create.
        """
        self.filesystem.makedirs(path=path, exist_ok=True)

    def mkdir(self, path: PathType) -> None:
        """Create a directory at the given path.

        Args:
            path: The path to create.
        """
        self.filesystem.makedir(path=path)

    def remove(self, path: PathType) -> None:
        """Remove the file at the given path.

        Args:
            path: The path of the file to remove.
        """
        self.filesystem.rm_file(path=path)

    def rename(
        self, src: PathType, dst: PathType, overwrite: bool = False
    ) -> None:
        """Rename source file to destination file.

        Args:
            src: The path of the file to rename.
            dst: The path to rename the source file to.
            overwrite: If a file already exists at the destination, this
                method will overwrite it if overwrite=`True` and
                raise a FileExistsError otherwise.

        Raises:
            FileExistsError: If a file already exists at the destination
                and overwrite is not set to `True`.
        """
        if not overwrite and self.filesystem.exists(dst):
            raise FileExistsError(
                f"Unable to rename file to '{convert_to_str(dst)}', "
                f"file already exists. Set `overwrite=True` to rename anyway."
            )

        # TODO [ENG-152]: Check if it works with overwrite=True or if we need
        #  to manually remove it first
        self.filesystem.rename(path1=src, path2=dst)

    def rmtree(self, path: PathType) -> None:
        """Remove the given directory.

        Args:
            path: The path of the directory to remove.
        """
        self.filesystem.delete(path=path, recursive=True)

    def stat(self, path: PathType) -> Dict[str, Any]:
        """Return stat info for the given path.

        Args:
            path: The path to get stat info for.

        Returns:
            A dictionary containing the stat info.
        """
        return self.filesystem.stat(path=path)  # type: ignore[no-any-return]

    def size(self, path: PathType) -> int:
        """Get the size of a file in bytes.

        Args:
            path: The path to the file.

        Returns:
            The size of the file in bytes.
        """
        return self.filesystem.size(path=path)  # type: ignore[no-any-return]

    def walk(
        self,
        top: PathType,
        topdown: bool = True,
        onerror: Optional[Callable[..., None]] = None,
    ) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        """Return an iterator that walks the contents of the given directory.

        Args:
            top: Path of directory to walk.
            topdown: Unused argument to conform to interface.
            onerror: Unused argument to conform to interface.

        Yields:
            An Iterable of Tuples, each of which contain the path of the current
                directory path, a list of directories inside the current directory
                and a list of files inside the current directory.
        """
        # TODO [ENG-153]: Additional params
        for directory, subdirectories, files in self.filesystem.walk(path=top):
            yield f"s3://{directory}", subdirectories, files

    def _remove_previous_file_versions(self, path: PathType) -> None:
        """Keep only the latest file version in the given path.

        Method is useful for logs stored in versioned file systems
        like AWS S3.

        Args:
            path: The path to the file.
        """
        if self.is_versioned:
            if isinstance(path, bytes):
                path = path.decode()
            _, prefix = split_s3_path(path)
            with self._shield_lack_of_versioning_permissions(
                "s3:ListBucketVersions"
            ):
                for version in self._boto3_bucket.object_versions.filter(
                    Prefix=prefix
                ):
                    if not version.is_latest:
                        with self._shield_lack_of_versioning_permissions(
                            "s3:DeleteObjectVersion"
                        ):
                            version.delete()

    def _build_boto3_kwargs(self) -> Dict[str, Any]:
        """Build boto3 kwargs by layering config overrides over connector credentials.

        Returns:
            A fresh dictionary of boto3 kwargs that callers can safely mutate.
        """
        key, secret, token, region = self.get_credentials()

        kwargs: Dict[str, Any] = (
            self.config.client_kwargs.copy()
            if self.config.client_kwargs
            else {}
        )

        if key is not None:
            kwargs["aws_access_key_id"] = key
        if secret is not None:
            kwargs["aws_secret_access_key"] = secret
        if token is not None:
            kwargs["aws_session_token"] = token
        if region is not None and "region_name" not in kwargs:
            kwargs["region_name"] = region

        return kwargs

    @property
    def _boto3_bucket(self) -> Any:
        """Get the boto3 bucket object.

        Returns:
            The boto3 bucket object.
        """
        if self._boto3_bucket_holder and not self.connector_has_expired():
            return self._boto3_bucket_holder

        s3 = boto3.resource("s3", **self._build_boto3_kwargs())
        self._boto3_bucket_holder = s3.Bucket(self.config.bucket)
        return self._boto3_bucket_holder

    @contextmanager
    def _shield_lack_of_versioning_permissions(
        self, auth_missing: str
    ) -> Generator[Any, None, None]:
        try:
            yield
        except ClientError as e:
            if "not authorized" in e.args[0] and auth_missing in e.args[0]:
                logger.warning(
                    "Your AWS Connector is lacking critical Versioning permissions. "
                    f"Please check that `{auth_missing}` is granted."
                    "This is needed to remove previous versions of log files from your "
                    "Artifact Store bucket."
                )
                self.is_versioned = False



"""Multiple and partial step inputs [d203788f82b9].

Revision ID: d203788f82b9
Revises: bc6317491676
Create Date: 2025-11-11 11:57:39.370022

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d203788f82b9"
down_revision = "bc6317491676"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision.

    Raises:
        NotImplementedError: If the database engine is not supported.
    """
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("item_count", sa.Integer(), nullable=True)
        )

    with op.batch_alter_table(
        "step_run_input_artifact", schema=None
    ) as batch_op:
        batch_op.add_column(
            sa.Column("input_index", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("chunk_index", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("chunk_size", sa.Integer(), nullable=True)
        )

    op.execute(sa.text("UPDATE step_run_input_artifact SET input_index=0"))
    with op.batch_alter_table(
        "step_run_input_artifact", schema=None
    ) as batch_op:
        batch_op.alter_column(
            "input_index",
            nullable=False,
            existing_type=sa.Integer(),
        )

    _disable_primary_key_requirement_if_necessary()

    meta = sa.MetaData()
    meta.reflect(bind=op.get_bind(), only=["step_run_input_artifact"])
    step_run_input_artifact_table = sa.Table("step_run_input_artifact", meta)

    pk_name = step_run_input_artifact_table.primary_key.name

    if not isinstance(pk_name, str):
        engine_name = op.get_bind().engine.name
        if engine_name == "sqlite":
            pk_name = "pk_step_run_input_artifact"
        elif engine_name in {"mysql", "mariadb"}:
            pk_name = "PRIMARY"
        else:
            raise NotImplementedError(f"Unsupported engine: {engine_name}")

    with op.batch_alter_table(
        "step_run_input_artifact",
        schema=None,
        naming_convention={"pk": "pk_%(table_name)s"},
    ) as batch_op:
        # Need to first drop the foreign keys as otherwise the table renaming
        # used during the PK constraint drop/create doesn't work
        batch_op.drop_constraint(
            constraint_name="fk_step_run_input_artifact_step_id_step_run",
            type_="foreignkey",
        )
        batch_op.drop_constraint(
            constraint_name="fk_step_run_input_artifact_artifact_id_artifact_version",
            type_="foreignkey",
        )

        # Update the PK
        batch_op.drop_constraint(constraint_name=pk_name, type_="primary")
        batch_op.create_primary_key(
            constraint_name="pk_step_run_input_artifact",
            columns=["step_id", "artifact_id", "name", "input_index"],
        )

        # Re-add the foreign keys
        batch_op.create_foreign_key(
            constraint_name="fk_step_run_input_artifact_step_id_step_run",
            referent_table="step_run",
            local_cols=["step_id"],
            remote_cols=["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            constraint_name="fk_step_run_input_artifact_artifact_id_artifact_version",
            referent_table="artifact_version",
            local_cols=["artifact_id"],
            remote_cols=["id"],
            ondelete="CASCADE",
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: Downgrade not possible.
    """
    raise NotImplementedError("Downgrade not possible")


def _disable_primary_key_requirement_if_necessary() -> None:
    """Adjusts settings based on database engine requirements.

    Raises:
        NotImplementedError: If the database engine is not supported.
    """
    conn = op.get_bind()
    engine = conn.engine
    engine_name = engine.name.lower()

    if engine_name == "mysql":
        server_version_info = engine.dialect.server_version_info
        if server_version_info is not None and server_version_info >= (
            8,
            0,
            13,
        ):
            potential_session_var = conn.execute(
                sa.text(
                    'SHOW SESSION VARIABLES LIKE "sql_require_primary_key";'
                )
            ).fetchone()
            if potential_session_var and potential_session_var[1] == "ON":
                # Temporarily disable this MySQL setting
                # for primary key modification
                op.execute("SET SESSION sql_require_primary_key = 0;")
    elif engine_name == "mariadb":
        # mariadb doesn't support session-based scoping for this setting
        op.execute("SET GLOBAL innodb_force_primary_key = 0;")
    elif engine_name == "sqlite":
        pass
    else:
        raise NotImplementedError(f"Unsupported engine: {engine_name}")



"""Pipeline source code [514b9fae1500].

Revision ID: 514b9fae1500
Revises: 2837a7ab3d77
Create Date: 2025-12-09 11:13:39.173696

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "514b9fae1500"
down_revision = "2837a7ab3d77"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.add_column(sa.Column("source_code", sa.TEXT(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.drop_column("source_code")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Containerized orchestrator class."""

from abc import ABC
from typing import List, Optional, Set

import zenml
from zenml.config.build_configuration import BuildConfiguration
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import ORCHESTRATOR_DOCKER_IMAGE_KEY
from zenml.models import PipelineSnapshotBase, PipelineSnapshotResponse
from zenml.orchestrators import BaseOrchestrator


class ContainerizedOrchestrator(BaseOrchestrator, ABC):
    """Base class for containerized orchestrators."""

    @property
    def requirements(self) -> Set[str]:
        """Set of PyPI requirements for the component.

        Returns:
            A set of PyPI requirements for the component.
        """
        requirements = super().requirements

        if self.config.is_local and GlobalConfiguration().uses_sql_store:
            # If we're directly connected to a DB, we need to install the
            # `local` extra in the Docker image to include the DB dependencies.
            requirements.add(f"zenml[local]=={zenml.__version__}")

        return requirements

    @staticmethod
    def get_image(
        snapshot: "PipelineSnapshotResponse",
        step_name: Optional[str] = None,
    ) -> str:
        """Gets the Docker image for the pipeline/a step.

        Args:
            snapshot: The snapshot from which to get the image.
            step_name: Pipeline step name for which to get the image. If not
                given the generic pipeline image will be returned.

        Raises:
            RuntimeError: If the snapshot does not have an associated build.

        Returns:
            The image name or digest.
        """
        if not snapshot.build:
            raise RuntimeError(
                f"Missing build for snapshot {snapshot.id}. This is "
                "probably because the build was manually deleted."
            )

        return snapshot.build.get_image(
            component_key=ORCHESTRATOR_DOCKER_IMAGE_KEY, step=step_name
        )

    def should_build_pipeline_image(
        self, snapshot: "PipelineSnapshotBase"
    ) -> bool:
        """Whether to build the pipeline image.

        Args:
            snapshot: The pipeline snapshot.

        Returns:
            Whether to build the pipeline image.
        """
        # When running a dynamic pipeline, we need an image for the
        # orchestration container.
        return snapshot.is_dynamic

    def get_docker_builds(
        self, snapshot: "PipelineSnapshotBase"
    ) -> List["BuildConfiguration"]:
        """Gets the Docker builds required for the component.

        Args:
            snapshot: The pipeline snapshot for which to get the builds.

        Returns:
            The required Docker builds.
        """
        pipeline_settings = snapshot.pipeline_configuration.docker_settings

        included_pipeline_build = False
        builds = []

        for name, step in snapshot.step_configurations.items():
            step_settings = step.config.docker_settings

            if step_settings != pipeline_settings:
                build = BuildConfiguration(
                    key=ORCHESTRATOR_DOCKER_IMAGE_KEY,
                    settings=step_settings,
                    step_name=name,
                )
                builds.append(build)
            elif not included_pipeline_build:
                pipeline_build = BuildConfiguration(
                    key=ORCHESTRATOR_DOCKER_IMAGE_KEY,
                    settings=pipeline_settings,
                )
                builds.append(pipeline_build)
                included_pipeline_build = True

        if not included_pipeline_build and self.should_build_pipeline_image(
            snapshot
        ):
            pipeline_build = BuildConfiguration(
                key=ORCHESTRATOR_DOCKER_IMAGE_KEY,
                settings=pipeline_settings,
            )
            builds.append(pipeline_build)

        return builds



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the LightGBM materializer."""

import os
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Tuple, Type

import lightgbm as lgb

from zenml.enums import ArtifactType
from zenml.io import fileio
from zenml.materializers.base_materializer import BaseMaterializer

if TYPE_CHECKING:
    from zenml.metadata.metadata_types import MetadataType

DEFAULT_FILENAME = "data.binary"


class LightGBMDatasetMaterializer(BaseMaterializer):
    """Materializer to read data to and from lightgbm.Dataset."""

    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (lgb.Dataset,)
    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA

    def load(self, data_type: Type[Any]) -> lgb.Dataset:
        """Reads a lightgbm.Dataset binary file and loads it.

        Args:
            data_type: A lightgbm.Dataset type.

        Returns:
            A lightgbm.Dataset object.
        """
        filepath = os.path.join(self.uri, DEFAULT_FILENAME)

        with self.get_temporary_directory(delete_at_exit=False) as temp_dir:
            temp_file = os.path.join(str(temp_dir), DEFAULT_FILENAME)

            # Copy from artifact store to temporary file
            fileio.copy(filepath, temp_file)
            matrix = lgb.Dataset(temp_file, free_raw_data=False)

            return matrix

    def save(self, matrix: lgb.Dataset) -> None:
        """Creates a binary serialization for a lightgbm.Dataset object.

        Args:
            matrix: A lightgbm.Dataset object.
        """
        filepath = os.path.join(self.uri, DEFAULT_FILENAME)

        with self.get_temporary_directory(delete_at_exit=True) as temp_dir:
            temp_file = os.path.join(str(temp_dir), DEFAULT_FILENAME)
            matrix.save_binary(temp_file)

            # Copy it into artifact store
            fileio.copy(temp_file, filepath)

    def extract_metadata(
        self, matrix: lgb.Dataset
    ) -> Dict[str, "MetadataType"]:
        """Extract metadata from the given `Dataset` object.

        Args:
            matrix: The `Dataset` object to extract metadata from.

        Returns:
            The extracted metadata as a dictionary.
        """
        return {"shape": (matrix.num_data(), matrix.num_feature())}



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Parameter group classes."""

from uuid import UUID

from pydantic import BaseModel, model_validator


class VersionedIdentifier(BaseModel):
    """Class grouping identifiers for entities resolved by UUID or name&version."""

    id: UUID | None = None
    name: str | None = None
    version: str | None = None

    @model_validator(mode="after")
    def _validate_options(self) -> "VersionedIdentifier":
        if self.id and self.name:
            raise ValueError(
                "You can use only identification option at a time."
                "Use either id or name."
            )

        if not (self.id or self.name):
            raise ValueError(
                "You have to use at least one identification option."
                "Use either id or name."
            )

        if bool(self.name) ^ bool(self.version):
            raise ValueError("You need to specify both name and version.")

        return self


class ArtifactVersionIdentifier(VersionedIdentifier):
    """Class for artifact version identifier group."""

    pass


class ModelVersionIdentifier(VersionedIdentifier):
    """Class for model version identifier group."""

    pass


class PipelineRunIdentifier(BaseModel):
    """Class grouping different pipeline run identifiers."""

    id: UUID | None = None
    name: str | None = None
    prefix: str | None = None

    @property
    def value(self) -> str | UUID:
        """Resolves the set value out of id, name, prefix etc.

        Returns:
            The id/name/prefix (if set, in this exact order).
        """
        return self.id or self.name or self.prefix  # type: ignore[return-value]

    @model_validator(mode="after")
    def _validate_options(self) -> "PipelineRunIdentifier":
        options = [
            bool(self.id),
            bool(self.name),
            bool(self.prefix),
        ]

        if sum(options) > 1:
            raise ValueError(
                "You can use only identification option at a time."
                "Use either id or name or prefix."
            )

        if sum(options) == 0:
            raise ValueError(
                "You have to use at least one identification option."
                "Use either id or name or prefix."
            )

        return self


class StepRunIdentifier(BaseModel):
    """Class grouping different step run identifiers."""

    id: UUID | None = None
    name: str | None = None
    run: PipelineRunIdentifier | None = None

    @model_validator(mode="after")
    def _validate_options(self) -> "StepRunIdentifier":
        if self.id and self.name:
            raise ValueError(
                "You can use only identification option at a time."
                "Use either id or name."
            )

        if not (self.id or self.name):
            raise ValueError(
                "You have to use at least one identification option."
                "Use either id or name."
            )

        if bool(self.name) ^ bool(self.run):
            raise ValueError(
                "To identify a step run by name you need to specify a pipeline run identifier."
            )

        return self



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Deployers are stack components responsible for deploying pipelines as HTTP services.

Deploying pipelines is the process of hosting and running machine-learning
pipelines as part of a managed web service and providing access to pipeline
execution through an API endpoint like HTTP or GRPC. Once deployed, you can send
execution requests to the pipeline through the web service's API and receive
responses containing the pipeline results or execution status.

Add a deployer to your ZenML stack to be able to provision pipelines deployments
that transform your ML pipelines into long-running HTTP services
for real-time, on-demand execution instead of traditional batch processing.

When present in a stack, the deployer also acts as a registry for pipeline
endpoints that are deployed with ZenML. You can use the deployer to list all
deployments that are currently provisioned for online execution or filtered
according to a particular snapshot or configuration, or to delete an external
deployment managed through ZenML.
"""

from zenml.deployers.base_deployer import (
    BaseDeployer,
    BaseDeployerFlavor,
    BaseDeployerConfig,
)
from zenml.deployers.containerized_deployer import (
    ContainerizedDeployer,
)
from zenml.deployers.docker.docker_deployer import (
    DockerDeployer,
    DockerDeployerConfig,
    DockerDeployerFlavor,
    DockerDeployerSettings,
)
from zenml.deployers.local.local_deployer import (
    LocalDeployer,
    LocalDeployerConfig,
    LocalDeployerFlavor,
    LocalDeployerSettings,
)

__all__ = [
    "BaseDeployer",
    "BaseDeployerFlavor",
    "BaseDeployerConfig",
    "ContainerizedDeployer",
    "DockerDeployer",
    "DockerDeployerConfig",
    "DockerDeployerFlavor",
    "DockerDeployerSettings",
    "LocalDeployer",
    "LocalDeployerConfig",
    "LocalDeployerFlavor",
    "LocalDeployerSettings",
]



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base class for all containerized deployers."""

from abc import ABC
from typing import (
    List,
    Set,
)

import zenml
from zenml.config.build_configuration import BuildConfiguration
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import (
    DEPLOYER_DOCKER_IMAGE_KEY,
)
from zenml.deployers.base_deployer import BaseDeployer
from zenml.deployers.utils import load_deployment_requirements
from zenml.logger import get_logger
from zenml.models import (
    PipelineSnapshotBase,
    PipelineSnapshotResponse,
)

logger = get_logger(__name__)


class ContainerizedDeployer(BaseDeployer, ABC):
    """Base class for all containerized deployers."""

    @staticmethod
    def get_image(snapshot: PipelineSnapshotResponse) -> str:
        """Get the docker image used to deploy a pipeline snapshot.

        Args:
            snapshot: The pipeline snapshot to get the image for.

        Returns:
            The docker image used to deploy the pipeline snapshot.

        Raises:
            RuntimeError: if the pipeline snapshot does not have a build or
                if the deployer image is not in the build.
        """
        if snapshot.build is None:
            raise RuntimeError("Pipeline snapshot does not have a build. ")
        if DEPLOYER_DOCKER_IMAGE_KEY not in snapshot.build.images:
            raise RuntimeError(
                "Pipeline snapshot build does not have a deployer image. "
            )
        return snapshot.build.images[DEPLOYER_DOCKER_IMAGE_KEY].image

    @property
    def requirements(self) -> Set[str]:
        """Set of PyPI requirements for the deployer.

        Returns:
            A set of PyPI requirements for the deployer.
        """
        requirements = super().requirements

        if self.config.is_local and GlobalConfiguration().uses_sql_store:
            # If we're directly connected to a DB, we need to install the
            # `local` extra in the Docker image to include the DB dependencies.
            requirements.add(f"zenml[local]=={zenml.__version__}")

        return requirements

    def get_docker_builds(
        self, snapshot: "PipelineSnapshotBase"
    ) -> List["BuildConfiguration"]:
        """Gets the Docker builds required for the component.

        Args:
            snapshot: The pipeline snapshot for which to get the builds.

        Returns:
            The required Docker builds.
        """
        deployment_settings = (
            snapshot.pipeline_configuration.deployment_settings
        )
        docker_settings = snapshot.pipeline_configuration.docker_settings
        if not docker_settings.install_deployment_requirements:
            return []

        deployment_requirements = load_deployment_requirements(
            deployment_settings
        )
        return [
            BuildConfiguration(
                key=DEPLOYER_DOCKER_IMAGE_KEY,
                settings=snapshot.pipeline_configuration.docker_settings,
                extra_requirements_files={
                    ".zenml_deployment_requirements": deployment_requirements,
                },
            )
        ]



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML deployers utilities."""

import json
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import jsonref
import requests

from zenml.client import Client
from zenml.config.deployment_settings import (
    DEFAULT_DEPLOYMENT_APP_INVOKE_URL_PATH,
    DeploymentSettings,
)
from zenml.config.step_configurations import Step
from zenml.deployers.exceptions import (
    DeploymentHTTPError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.enums import DeploymentStatus, StepRuntime
from zenml.models import (
    CodeReferenceRequest,
    DeploymentResponse,
    PipelineSnapshotRequest,
    PipelineSnapshotResponse,
)
from zenml.steps.step_context import get_step_context
from zenml.utils import pydantic_utils
from zenml.utils.json_utils import pydantic_encoder


def get_deployment_input_schema(
    deployment: DeploymentResponse,
) -> Dict[str, Any]:
    """Get the schema for a deployment's input parameters.

    Args:
        deployment: The deployment for which to get the schema.

    Returns:
        The schema for the deployment's input parameters.

    Raises:
        RuntimeError: If the deployment has no associated input schema.
    """
    if (
        deployment.snapshot
        and deployment.snapshot.pipeline_spec
        and deployment.snapshot.pipeline_spec.input_schema
    ):
        return deployment.snapshot.pipeline_spec.input_schema

    raise RuntimeError(
        f"Deployment {deployment.name} has no associated input schema."
    )


def get_deployment_output_schema(
    deployment: DeploymentResponse,
) -> Dict[str, Any]:
    """Get the schema for a deployment's output parameters.

    Args:
        deployment: The deployment for which to get the schema.

    Returns:
        The schema for the deployment's output parameters.

    Raises:
        RuntimeError: If the deployment has no associated output schema.
    """
    if (
        deployment.snapshot
        and deployment.snapshot.pipeline_spec
        and deployment.snapshot.pipeline_spec.output_schema
    ):
        return deployment.snapshot.pipeline_spec.output_schema

    raise RuntimeError(
        f"Deployment {deployment.name} has no associated output schema."
    )


def get_deployment_invocation_example(
    deployment: DeploymentResponse,
) -> Dict[str, Any]:
    """Generate an example invocation command for a deployment.

    Args:
        deployment: The deployment for which to generate an example invocation.

    Returns:
        A dictionary containing the example invocation parameters.
    """
    parameters_schema = get_deployment_input_schema(deployment)

    properties = parameters_schema.get("properties", {})

    if not properties:
        return {}

    parameters = {}

    for attr_name, attr_schema in properties.items():
        parameters[attr_name] = "<value>"
        if not isinstance(attr_schema, dict):
            continue

        default_value = None

        if "default" in attr_schema:
            default_value = attr_schema["default"]
        elif "const" in attr_schema:
            default_value = attr_schema["const"]

        parameters[attr_name] = default_value or "<value>"

    return parameters


def invoke_deployment(
    deployment_name_or_id: Union[str, UUID],
    project: Optional[UUID] = None,
    timeout: int = 300,  # 5 minute timeout
    **kwargs: Any,
) -> Any:
    """Call a deployment and return the result.

    Args:
        deployment_name_or_id: The name or ID of the deployment to call.
        project: The project ID of the deployment to call.
        timeout: The timeout for the HTTP request to the deployment.
        **kwargs: Keyword arguments to pass to the deployment.

    Returns:
        The response from the deployment, parsed as JSON if possible,
        otherwise returned as text.

    Raises:
        DeploymentNotFoundError: If the deployment is not found.
        DeploymentProvisionError: If the deployment is not running
            or has no URL.
        DeploymentHTTPError: If the HTTP request to the endpoint fails.
    """
    client = Client()
    try:
        deployment = client.get_deployment(
            deployment_name_or_id, project=project
        )
    except KeyError:
        raise DeploymentNotFoundError(
            f"Deployment with name or ID '{deployment_name_or_id}' not found"
        )

    if deployment.status != DeploymentStatus.RUNNING:
        raise DeploymentProvisionError(
            f"Deployment {deployment_name_or_id} is not running. Please "
            "refresh or re-deploy the deployment or check its logs for "
            "more details."
        )

    if not deployment.url:
        raise DeploymentProvisionError(
            f"Deployment {deployment_name_or_id} has no URL. Please "
            "refresh the deployment or check its logs for more "
            "details."
        )

    input_schema = None
    if deployment.snapshot and deployment.snapshot.pipeline_spec:
        input_schema = deployment.snapshot.pipeline_spec.input_schema

    if input_schema:
        # Resolve the references in the schema first, otherwise we won't be able
        # to access the data types for object-typed parameters.
        input_schema = jsonref.replace_refs(input_schema)
        assert isinstance(input_schema, dict)

        properties = input_schema.get("properties", {})

        # Some kwargs having one of the collection data types (list, dict) in
        # the schema may be supplied as a JSON string. We need to unpack
        # them before we construct the final JSON payload.
        #
        # We ignore all errors here because they will be better handled by the
        # deployment itself server side.
        for key in kwargs.keys():
            if key not in properties:
                continue
            value = kwargs[key]
            if not isinstance(value, str):
                continue
            attr_schema = properties[key]
            try:
                if attr_schema.get("type") == "object":
                    value = json.loads(value)
                    if isinstance(value, dict):
                        kwargs[key] = value
                elif attr_schema.get("type") == "array":
                    value = json.loads(value)
                    if isinstance(value, list):
                        kwargs[key] = value
            except (json.JSONDecodeError, ValueError):
                pass

    # Serialize kwargs to JSON
    params = dict(parameters=kwargs)
    try:
        payload = json.dumps(params, default=pydantic_encoder)
    except (TypeError, ValueError) as e:
        raise DeploymentHTTPError(
            f"Failed to serialize request data to JSON: {e}"
        )

    invoke_url_path = DEFAULT_DEPLOYMENT_APP_INVOKE_URL_PATH
    if deployment.snapshot:
        deployment_settings = (
            deployment.snapshot.pipeline_configuration.deployment_settings
        )
        invoke_url_path = f"{deployment_settings.root_url_path}{deployment_settings.api_url_path}{deployment_settings.invoke_url_path}"

    # Construct the invoke endpoint URL
    invoke_url = deployment.url.rstrip("/") + invoke_url_path

    # Prepare headers
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    # Add authorization header if auth_key is present
    if deployment.auth_key:
        headers["Authorization"] = f"Bearer {deployment.auth_key}"

    try:
        step_context = get_step_context()
    except RuntimeError:
        step_context = None

    if step_context:
        # Include these so that the deployment can identify the step
        # and pipeline run that called it, if called from a step.
        headers["ZenML-Step-Name"] = step_context.step_name
        headers["ZenML-Pipeline-Name"] = step_context.pipeline.name
        headers["ZenML-Pipeline-Run-ID"] = str(step_context.pipeline_run.id)
        headers["ZenML-Pipeline-Run-Name"] = step_context.pipeline_run.name

    # Make the HTTP request
    try:
        response = requests.post(
            invoke_url,
            data=payload,
            headers=headers,
            timeout=timeout,
        )
        response.raise_for_status()

        # Try to parse JSON response, fallback to text if not JSON
        try:
            return response.json()
        except ValueError:
            return response.text

    except requests.exceptions.HTTPError as e:
        raise DeploymentHTTPError(
            f"HTTP {e.response.status_code} error calling deployment "
            f"{deployment_name_or_id}: {e.response.text}"
        )
    except requests.exceptions.ConnectionError as e:
        raise DeploymentHTTPError(
            f"Failed to connect to deployment {deployment_name_or_id}: {e}"
        )
    except requests.exceptions.Timeout as e:
        raise DeploymentHTTPError(
            f"Timeout calling deployment {deployment_name_or_id}: {e}"
        )
    except requests.exceptions.RequestException as e:
        raise DeploymentHTTPError(
            f"Request failed for deployment {deployment_name_or_id}: {e}"
        )


def deployment_snapshot_request_from_source_snapshot(
    source_snapshot: PipelineSnapshotResponse,
    deployment_parameters: Dict[str, Any],
) -> PipelineSnapshotRequest:
    """Generate a snapshot request for deployment execution.

    Args:
        source_snapshot: The source snapshot from which to create the
            snapshot request.
        deployment_parameters: Parameters to override for deployment execution.

    Raises:
        RuntimeError: If the source snapshot does not have an associated stack.

    Returns:
        The generated snapshot request.
    """
    if source_snapshot.stack is None:
        raise RuntimeError("Missing source snapshot stack")

    pipeline_configuration = pydantic_utils.update_model(
        source_snapshot.pipeline_configuration, {"enable_cache": False}
    )

    steps = {}
    for invocation_id, step in source_snapshot.step_configurations.items():
        updated_step_parameters = step.config.parameters.copy()

        for param_name in step.config.parameters:
            if param_name in deployment_parameters:
                updated_step_parameters[param_name] = deployment_parameters[
                    param_name
                ]

        # Deployment-specific step overrides
        step_update = {
            "enable_cache": False,  # Disable caching for all steps
            "step_operator": None,  # Remove step operators for deployments
            "parameters": updated_step_parameters,
            "runtime": StepRuntime.INLINE,
        }

        step_config = pydantic_utils.update_model(
            step.step_config_overrides, step_update
        )
        merged_step_config = step_config.apply_pipeline_configuration(
            pipeline_configuration
        )

        step_spec = step.spec.model_copy(update={"enable_heartbeat": False})

        steps[invocation_id] = Step(
            spec=step_spec,
            config=merged_step_config,
            step_config_overrides=step_config,
        )

    code_reference_request = None
    if source_snapshot.code_reference:
        code_reference_request = CodeReferenceRequest(
            commit=source_snapshot.code_reference.commit,
            subdirectory=source_snapshot.code_reference.subdirectory,
            code_repository=source_snapshot.code_reference.code_repository.id,
        )

    zenml_version = Client().zen_store.get_store_info().version

    # Compute the source snapshot ID:
    # - If the source snapshot has a name, we use it as the source snapshot.
    #   That way, all runs will be associated with this snapshot.
    # - If the source snapshot is based on another snapshot (which therefore
    #   has a name), we use that one instead.
    # - If the source snapshot does not have a name and is not based on another
    #   snapshot, we don't set a source snapshot.
    #
    # With this, we ensure that all runs are associated with the closest named
    # source snapshot.
    source_snapshot_id = None
    if source_snapshot.name:
        source_snapshot_id = source_snapshot.id
    elif source_snapshot.source_snapshot_id:
        source_snapshot_id = source_snapshot.source_snapshot_id

    updated_pipeline_spec = source_snapshot.pipeline_spec
    if (
        source_snapshot.pipeline_spec
        and source_snapshot.pipeline_spec.parameters is not None
    ):
        original_params: Dict[str, Any] = dict(
            source_snapshot.pipeline_spec.parameters
        )
        merged_params: Dict[str, Any] = original_params.copy()
        for k, v in deployment_parameters.items():
            if k in original_params:
                merged_params[k] = v
        updated_pipeline_spec = pydantic_utils.update_model(
            source_snapshot.pipeline_spec, {"parameters": merged_params}
        )

    return PipelineSnapshotRequest(
        project=source_snapshot.project_id,
        run_name_template=source_snapshot.run_name_template,
        pipeline_configuration=pipeline_configuration,
        step_configurations=steps,
        client_environment={},
        client_version=zenml_version,
        server_version=zenml_version,
        stack=source_snapshot.stack.id,
        pipeline=source_snapshot.pipeline.id,
        schedule=None,
        code_reference=code_reference_request,
        code_path=source_snapshot.code_path,
        build=source_snapshot.build.id if source_snapshot.build else None,
        source_snapshot=source_snapshot_id,
        pipeline_version_hash=source_snapshot.pipeline_version_hash,
        pipeline_spec=updated_pipeline_spec,
        source_code=source_snapshot.source_code,
        is_dynamic=source_snapshot.is_dynamic,
    )


def load_deployment_requirements(
    deployment_settings: DeploymentSettings,
) -> List[str]:
    """Load the software requirements for a deployment.

    Args:
        deployment_settings: The deployment settings for which to load the
            software requirements.

    Returns:
        The software requirements for the deployment.

    Raises:
        RuntimeError: If the deployment app runner flavor cannot be loaded.
    """
    from zenml.deployers.server.app import BaseDeploymentAppRunnerFlavor

    try:
        deployment_app_runner_flavor = (
            BaseDeploymentAppRunnerFlavor.load_app_runner_flavor(
                deployment_settings
            )
        )
    except Exception as e:
        raise RuntimeError(
            f"Failed to load deployment app runner flavor from deployment "
            f"settings: {e}"
        ) from e

    return deployment_app_runner_flavor.requirements



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation for Seldon secret schemas."""

from typing import Optional

from typing_extensions import Literal

from zenml.secret.base_secret import BaseSecretSchema


class SeldonS3SecretSchema(BaseSecretSchema):
    """Seldon S3 credentials.

    Based on: https://rclone.org/s3/#amazon-s3

    Attributes:
        rclone_config_s3_type: the rclone config type. Must be set to "s3" for
            this schema.
        rclone_config_s3_provider: the S3 provider (e.g. aws, ceph, minio).
        rclone_config_s3_env_auth: get AWS credentials from EC2/ECS meta data
            (i.e. with IAM roles configuration). Only applies if access_key_id
            and secret_access_key are blank.
        rclone_config_s3_access_key_id: AWS Access Key ID.
        rclone_config_s3_secret_access_key: AWS Secret Access Key.
        rclone_config_s3_session_token: AWS Session Token.
        rclone_config_s3_region: region to connect to.
        rclone_config_s3_endpoint: S3 API endpoint.

    """

    rclone_config_s3_type: Literal["s3"] = "s3"
    rclone_config_s3_provider: str = "aws"
    rclone_config_s3_env_auth: bool = False
    rclone_config_s3_access_key_id: Optional[str] = None
    rclone_config_s3_secret_access_key: Optional[str] = None
    rclone_config_s3_session_token: Optional[str] = None
    rclone_config_s3_region: Optional[str] = None
    rclone_config_s3_endpoint: Optional[str] = None


class SeldonGSSecretSchema(BaseSecretSchema):
    """Seldon GCS credentials.

    Based on: https://rclone.org/googlecloudstorage/

    Attributes:
        rclone_config_gs_type: the rclone config type. Must be set to "google
            cloud storage" for this schema.
        rclone_config_gs_client_id: OAuth client id.
        rclone_config_gs_client_secret: OAuth client secret.
        rclone_config_gs_token: OAuth Access Token as a JSON blob.
        rclone_config_gs_project_number: project number.
        rclone_config_gs_service_account_credentials: service account
            credentials JSON blob.
        rclone_config_gs_anonymous: access public buckets and objects without
            credentials. Set to True if you just want to download files and
            don't configure credentials.
        rclone_config_gs_auth_url: auth server URL.
    """

    rclone_config_gs_type: Literal["google cloud storage"] = (
        "google cloud storage"
    )
    rclone_config_gs_client_id: Optional[str] = None
    rclone_config_gs_client_secret: Optional[str] = None
    rclone_config_gs_project_number: Optional[str] = None
    rclone_config_gs_service_account_credentials: Optional[str] = None
    rclone_config_gs_anonymous: bool = False
    rclone_config_gs_token: Optional[str] = None
    rclone_config_gs_auth_url: Optional[str] = None
    rclone_config_gs_token_url: Optional[str] = None


class SeldonAzureSecretSchema(BaseSecretSchema):
    """Seldon Azure Blob Storage credentials.

    Based on: https://rclone.org/azureblob/

    Attributes:
        rclone_config_az_type: the rclone config type. Must be set to
            "azureblob" for this schema.
        rclone_config_az_env_auth: read credentials from runtime
            (environment variables or MSI).
        rclone_config_az_account: storage Account Name. Leave blank to
            use SAS URL or MSI.
        rclone_config_az_key: storage Account Key. Leave blank to
            use SAS URL or MSI.
        rclone_config_az_sas_url: SAS URL for container level access
            only. Leave blank if using account/key or MSI.
        rclone_config_az_use_msi: use a managed service identity to
            authenticate (only works in Azure).
        rclone_config_az_client_secret: client secret for service
            principal authentication.
        rclone_config_az_client_id: client id for service principal
            authentication.
        rclone_config_az_tenant: tenant id for service principal
            authentication.
    """

    rclone_config_az_type: Literal["azureblob"] = "azureblob"
    rclone_config_az_env_auth: bool = False
    rclone_config_az_account: Optional[str] = None
    rclone_config_az_key: Optional[str] = None
    rclone_config_az_sas_url: Optional[str] = None
    rclone_config_az_use_msi: bool = False
    rclone_config_az_client_secret: Optional[str] = None
    rclone_config_az_client_id: Optional[str] = None
    rclone_config_az_tenant: Optional[str] = None



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Secure headers for the ZenML Server."""

from typing import Optional

import secure

from zenml.zen_server.utils import server_config

_secure_headers: Optional[secure.Secure] = None


def secure_headers() -> secure.Secure:
    """Return the secure headers component.

    Returns:
        The secure headers component.

    Raises:
        RuntimeError: If the secure headers component is not initialized.
    """
    global _secure_headers
    if _secure_headers is None:
        raise RuntimeError("Secure headers component not initialized")
    return _secure_headers


def initialize_secure_headers() -> None:
    """Initialize the secure headers component."""
    global _secure_headers

    config = server_config()

    # For each of the secure headers supported by the `secure` library, we
    # check if the corresponding configuration is set in the server
    # configuration:
    #
    # - if set to `True`, we use the default value for the header
    # - if set to a string, we use the string as the value for the header
    # - if set to `False`, we don't set the header

    server: Optional[secure.Server] = None
    if config.secure_headers_server:
        server = secure.Server()
        if isinstance(config.secure_headers_server, str):
            server.set(config.secure_headers_server)
        else:
            server.set(str(config.deployment_id))

    hsts: Optional[secure.StrictTransportSecurity] = None
    if config.secure_headers_hsts:
        hsts = secure.StrictTransportSecurity()
        if isinstance(config.secure_headers_hsts, str):
            hsts.set(config.secure_headers_hsts)

    xfo: Optional[secure.XFrameOptions] = None
    if config.secure_headers_xfo:
        xfo = secure.XFrameOptions()
        if isinstance(config.secure_headers_xfo, str):
            xfo.set(config.secure_headers_xfo)

    csp: Optional[secure.ContentSecurityPolicy] = None
    if config.secure_headers_csp:
        csp = secure.ContentSecurityPolicy()
        if isinstance(config.secure_headers_csp, str):
            csp.set(config.secure_headers_csp)

    xcto: Optional[secure.XContentTypeOptions] = None
    if config.secure_headers_content:
        xcto = secure.XContentTypeOptions()
        if isinstance(config.secure_headers_content, str):
            xcto.set(config.secure_headers_content)

    referrer: Optional[secure.ReferrerPolicy] = None
    if config.secure_headers_referrer:
        referrer = secure.ReferrerPolicy()
        if isinstance(config.secure_headers_referrer, str):
            referrer.set(config.secure_headers_referrer)

    cache: Optional[secure.CacheControl] = None
    if config.secure_headers_cache:
        cache = secure.CacheControl()
        if isinstance(config.secure_headers_cache, str):
            cache.set(config.secure_headers_cache)

    permissions: Optional[secure.PermissionsPolicy] = None
    if config.secure_headers_permissions:
        permissions = secure.PermissionsPolicy()
        if isinstance(config.secure_headers_permissions, str):
            # This one is special, because it doesn't allow setting the
            # value as a string, but rather as a list of directives, so we
            # hack our way around it by setting the private _default_value
            # attribute.
            permissions._default_value = config.secure_headers_permissions

    _secure_headers = secure.Secure(
        server=server,
        hsts=hsts,
        xfo=xfo,
        csp=csp,
        xcto=xcto,
        referrer=referrer,
        cache=cache,
        permissions=permissions,
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for run metadata."""

from typing import Any, List, Optional, Union
from uuid import UUID

from fastapi import APIRouter, Security

from zenml.constants import API, RUN_METADATA, VERSION_1
from zenml.enums import MetadataResourceTypes
from zenml.models import RunMetadataRequest
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.models import Action
from zenml.zen_server.rbac.utils import (
    batch_verify_permissions_for_models,
    verify_permission_for_model,
)
from zenml.zen_server.routers.projects_endpoints import workspace_router
from zenml.zen_server.utils import async_fastapi_endpoint_wrapper, zen_store

router = APIRouter(
    prefix=API + VERSION_1 + RUN_METADATA,
    tags=["run_metadata"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.post(
    "/{project_name_or_id}" + RUN_METADATA,
    responses={401: error_response, 409: error_response, 422: error_response},
    deprecated=True,
    tags=["run_metadata"],
)
@async_fastapi_endpoint_wrapper
def create_run_metadata(
    run_metadata: RunMetadataRequest,
    project_name_or_id: Optional[Union[str, UUID]] = None,
    auth_context: AuthContext = Security(authorize),
) -> None:
    """Creates run metadata.

    Args:
        run_metadata: The run metadata to create.
        project_name_or_id: Optional name or ID of the project.
        auth_context: Authentication context.

    Raises:
        RuntimeError: If the resource type is not supported.
    """
    if project_name_or_id:
        project = zen_store().get_project(project_name_or_id)
        run_metadata.project = project.id

    run_metadata.user = auth_context.user.id

    verify_models: List[Any] = []
    for resource in run_metadata.resources:
        if resource.type == MetadataResourceTypes.PIPELINE_RUN:
            verify_models.append(zen_store().get_run(resource.id))
        elif resource.type == MetadataResourceTypes.STEP_RUN:
            verify_models.append(zen_store().get_run_step(resource.id))
        elif resource.type == MetadataResourceTypes.ARTIFACT_VERSION:
            verify_models.append(zen_store().get_artifact_version(resource.id))
        elif resource.type == MetadataResourceTypes.MODEL_VERSION:
            verify_models.append(zen_store().get_model_version(resource.id))
        elif resource.type == MetadataResourceTypes.SCHEDULE:
            verify_models.append(zen_store().get_schedule(resource.id))
        elif resource.type == MetadataResourceTypes.WAIT_CONDITION:
            wait_condition = zen_store().get_run_wait_condition(resource.id)
            verify_models.append(wait_condition.run)
        else:
            raise RuntimeError(f"Unknown resource type: {resource.type}")

    batch_verify_permissions_for_models(
        models=verify_models,
        action=Action.UPDATE,
    )

    verify_permission_for_model(model=run_metadata, action=Action.CREATE)

    zen_store().create_run_metadata(run_metadata)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the AWS App Runner deployer."""

import datetime
import json
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

import boto3
from botocore.exceptions import BotoCoreError, ClientError
from pydantic import BaseModel

from zenml.config.base_settings import BaseSettings
from zenml.config.resource_settings import ResourceSettings
from zenml.deployers.containerized_deployer import ContainerizedDeployer
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentDeprovisionError,
    DeploymentLogsNotFoundError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.deployers.server.entrypoint_configuration import (
    DEPLOYMENT_ID_OPTION,
    DeploymentEntrypointConfiguration,
)
from zenml.enums import DeploymentStatus, StackComponentType
from zenml.integrations.aws.flavors.aws_deployer_flavor import (
    AWSDeployerConfig,
    AWSDeployerSettings,
)
from zenml.logger import get_logger
from zenml.models import (
    DeploymentOperationalState,
    DeploymentResponse,
)
from zenml.stack import StackValidator

if TYPE_CHECKING:
    from zenml.stack import Stack

logger = get_logger(__name__)

# Default resource and scaling configuration constants
DEFAULT_CPU = 0.25  # vCPU
DEFAULT_MEMORY = 0.5  # GB
DEFAULT_MIN_REPLICAS = 1
DEFAULT_MAX_REPLICAS = 25
DEFAULT_MAX_CONCURRENCY = 100

# AWS App Runner built-in limits
AWS_APP_RUNNER_MAX_SIZE = 1000
AWS_APP_RUNNER_MAX_CONCURRENCY = 1000


class AppRunnerDeploymentMetadata(BaseModel):
    """Metadata for an App Runner deployment."""

    service_name: Optional[str] = None
    service_arn: Optional[str] = None
    service_url: Optional[str] = None
    region: Optional[str] = None
    service_id: Optional[str] = None
    status: Optional[str] = None
    source_configuration: Optional[Dict[str, Any]] = None
    instance_configuration: Optional[Dict[str, Any]] = None
    auto_scaling_configuration_summary: Optional[Dict[str, Any]] = None
    auto_scaling_configuration_arn: Optional[str] = None
    health_check_configuration: Optional[Dict[str, Any]] = None
    network_configuration: Optional[Dict[str, Any]] = None
    observability_configuration: Optional[Dict[str, Any]] = None
    encryption_configuration: Optional[Dict[str, Any]] = None
    cpu: Optional[str] = None
    memory: Optional[str] = None
    port: Optional[int] = None
    auto_scaling_max_concurrency: Optional[int] = None
    auto_scaling_max_size: Optional[int] = None
    auto_scaling_min_size: Optional[int] = None
    is_publicly_accessible: Optional[bool] = None
    health_check_grace_period_seconds: Optional[int] = None
    health_check_interval_seconds: Optional[int] = None
    health_check_path: Optional[str] = None
    health_check_protocol: Optional[str] = None
    health_check_timeout_seconds: Optional[int] = None
    health_check_healthy_threshold: Optional[int] = None
    health_check_unhealthy_threshold: Optional[int] = None
    tags: Optional[Dict[str, str]] = None
    traffic_allocation: Optional[Dict[str, int]] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    deleted_at: Optional[str] = None
    secret_arn: Optional[str] = None

    @classmethod
    def from_app_runner_service(
        cls,
        service: Dict[str, Any],
        region: str,
        secret_arn: Optional[str] = None,
    ) -> "AppRunnerDeploymentMetadata":
        """Create metadata from an App Runner service.

        Args:
            service: The App Runner service dictionary from describe_service.
            region: The AWS region.
            secret_arn: The AWS Secrets Manager secret ARN for the deployment.

        Returns:
            The metadata for the App Runner service.
        """
        instance_config = service.get("InstanceConfiguration", {})
        cpu = instance_config.get("Cpu")
        memory = instance_config.get("Memory")

        auto_scaling_config = service.get(
            "AutoScalingConfigurationSummary", {}
        )
        auto_scaling_configuration_arn = auto_scaling_config.get(
            "AutoScalingConfigurationArn"
        )
        auto_scaling_max_concurrency = auto_scaling_config.get(
            "MaxConcurrency"
        )
        auto_scaling_max_size = auto_scaling_config.get("MaxSize")
        auto_scaling_min_size = auto_scaling_config.get("MinSize")

        health_check_config = service.get("HealthCheckConfiguration", {})
        health_check_grace_period = health_check_config.get(
            "HealthCheckGracePeriodSeconds"
        )
        health_check_interval = health_check_config.get("Interval")
        health_check_path = health_check_config.get("Path")
        health_check_protocol = health_check_config.get("Protocol")
        health_check_timeout = health_check_config.get("Timeout")
        health_check_healthy_threshold = health_check_config.get(
            "HealthyThreshold"
        )
        health_check_unhealthy_threshold = health_check_config.get(
            "UnhealthyThreshold"
        )

        network_config = service.get("NetworkConfiguration", {})
        is_publicly_accessible = network_config.get(
            "IngressConfiguration", {}
        ).get("IsPubliclyAccessible")

        source_config = service.get("SourceConfiguration", {})
        image_repo = source_config.get("ImageRepository", {})
        image_config = image_repo.get("ImageConfiguration", {})

        port = None
        if image_config:
            port = image_config.get("Port")

        traffic_allocation = {}
        traffic_config = service.get("TrafficConfiguration", [])
        for traffic in traffic_config:
            if traffic.get("Type") == "LATEST":
                traffic_allocation["LATEST"] = traffic.get("Percent", 0)
            elif traffic.get("Revision"):
                traffic_allocation[traffic["Revision"]] = traffic.get(
                    "Percent", 0
                )
            elif traffic.get("Tag"):
                traffic_allocation[f"tag:{traffic['Tag']}"] = traffic.get(
                    "Percent", 0
                )

        # Extract timestamps
        created_at = service.get("CreatedAt")
        updated_at = service.get("UpdatedAt")
        deleted_at = service.get("DeletedAt")

        return cls(
            service_name=service.get("ServiceName"),
            service_arn=service.get("ServiceArn"),
            service_url=service.get("ServiceUrl"),
            region=region,
            service_id=service.get("ServiceId"),
            status=service.get("Status"),
            source_configuration=source_config,
            instance_configuration=instance_config,
            auto_scaling_configuration_summary=auto_scaling_config,
            auto_scaling_configuration_arn=auto_scaling_configuration_arn,
            health_check_configuration=health_check_config,
            network_configuration=network_config,
            observability_configuration=service.get(
                "ObservabilityConfiguration"
            ),
            encryption_configuration=service.get("EncryptionConfiguration"),
            cpu=cpu,
            memory=memory,
            port=port,
            auto_scaling_max_concurrency=auto_scaling_max_concurrency,
            auto_scaling_max_size=auto_scaling_max_size,
            auto_scaling_min_size=auto_scaling_min_size,
            is_publicly_accessible=is_publicly_accessible,
            health_check_grace_period_seconds=health_check_grace_period,
            health_check_interval_seconds=health_check_interval,
            health_check_path=health_check_path,
            health_check_protocol=health_check_protocol,
            health_check_timeout_seconds=health_check_timeout,
            health_check_healthy_threshold=health_check_healthy_threshold,
            health_check_unhealthy_threshold=health_check_unhealthy_threshold,
            tags=dict(service.get("Tags", {})),
            traffic_allocation=traffic_allocation
            if traffic_allocation
            else None,
            created_at=created_at.isoformat() if created_at else None,
            updated_at=updated_at.isoformat() if updated_at else None,
            deleted_at=deleted_at.isoformat() if deleted_at else None,
            secret_arn=secret_arn,
        )

    @classmethod
    def from_deployment(
        cls, deployment: DeploymentResponse
    ) -> "AppRunnerDeploymentMetadata":
        """Create metadata from a deployment.

        Args:
            deployment: The deployment to get the metadata for.

        Returns:
            The metadata for the deployment.
        """
        return cls.model_validate(deployment.deployment_metadata)


class AWSDeployer(ContainerizedDeployer):
    """Deployer responsible for deploying pipelines on AWS App Runner."""

    _boto_session: Optional[boto3.Session] = None
    _region: Optional[str] = None
    _app_runner_client: Optional[Any] = None
    _secrets_manager_client: Optional[Any] = None
    _logs_client: Optional[Any] = None

    @property
    def config(self) -> AWSDeployerConfig:
        """Returns the `AWSDeployerConfig` config.

        Returns:
            The configuration.
        """
        return cast(AWSDeployerConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the AWS deployer.

        Returns:
            The settings class.
        """
        return AWSDeployerSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures there is an image builder in the stack.

        Returns:
            A `StackValidator` instance.
        """
        return StackValidator(
            required_components={
                StackComponentType.IMAGE_BUILDER,
                StackComponentType.CONTAINER_REGISTRY,
            }
        )

    def _get_boto_session_and_region(self) -> Tuple[boto3.Session, str]:
        """Get an authenticated boto3 session and determine the region.

        Returns:
            A tuple containing the boto3 session and the AWS region.

        Raises:
            RuntimeError: If the service connector returns an unexpected type.
        """
        if (
            self._boto_session is not None
            and self._region is not None
            and not self.connector_has_expired()
        ):
            return self._boto_session, self._region

        # Option 1: Service connector
        if connector := self.get_connector():
            boto_session = connector.connect()
            if not isinstance(boto_session, boto3.Session):
                raise RuntimeError(
                    f"Expected to receive a `boto3.Session` object from the "
                    f"linked connector, but got type `{type(boto_session)}`."
                )

            region = boto_session.region_name
            if not region:
                # Fallback to config region or default
                region = self.config.region or "us-east-1"
                logger.warning(
                    f"No region found in boto3 session, using {region}"
                )

        # Option 2: Implicit configuration
        else:
            boto_session = boto3.Session(region_name=self.config.region)

        self._boto_session = boto_session
        self._region = region
        return boto_session, region

    @property
    def app_runner_client(self) -> Any:
        """Get the App Runner client.

        Returns:
            The App Runner client.
        """
        if self._app_runner_client is None or self.connector_has_expired():
            session, region = self._get_boto_session_and_region()
            self._app_runner_client = session.client(
                "apprunner", region_name=region
            )
        return self._app_runner_client

    @property
    def secrets_manager_client(self) -> Any:
        """Get the Secrets Manager client.

        Returns:
            The Secrets Manager client.
        """
        if (
            self._secrets_manager_client is None
            or self.connector_has_expired()
        ):
            session, region = self._get_boto_session_and_region()
            self._secrets_manager_client = session.client(
                "secretsmanager", region_name=region
            )
        return self._secrets_manager_client

    @property
    def logs_client(self) -> Any:
        """Get the CloudWatch Logs client.

        Returns:
            The CloudWatch Logs client.
        """
        if self._logs_client is None or self.connector_has_expired():
            session, region = self._get_boto_session_and_region()
            self._logs_client = session.client("logs", region_name=region)
        return self._logs_client

    @property
    def region(self) -> str:
        """Get the AWS region.

        Returns:
            The AWS region.
        """
        _, region = self._get_boto_session_and_region()
        return region

    def get_tags(
        self,
        deployment: DeploymentResponse,
        settings: AWSDeployerSettings,
    ) -> List[Dict[str, str]]:
        """Get the tags for a deployment to be used for AWS resources.

        Args:
            deployment: The deployment.
            settings: The deployer settings.

        Returns:
            The tags for the deployment.
        """
        tags = {
            **settings.tags,
            "zenml-deployment-id": str(deployment.id),
            "zenml-deployment-name": deployment.name,
            "zenml-deployer-name": str(self.name),
            "zenml-deployer-id": str(self.id),
            "managed-by": "zenml",
        }

        return [{"Key": k, "Value": v} for k, v in tags.items()]

    def _sanitize_name(
        self,
        name: str,
        random_suffix: str,
        max_length: int = 32,
        extra_allowed_characters: str = "-_",
    ) -> str:
        """Sanitize a name to comply with AWS naming requirements.

        Common AWS naming requirements:
        - Length: 4-max_length characters
        - Characters: letters (a-z, A-Z), numbers (0-9), configured extra
        allowed characters (e.g. dashes and underscores)
        - Must start and end with a letter or number
        - Cannot contain consecutive extra_allowed_characters

        Args:
            name: The raw name to sanitize.
            random_suffix: A random suffix to add to the name to ensure
                uniqueness.
            max_length: The maximum length of the name.
            extra_allowed_characters: Extra allowed characters in the name.

        Returns:
            A sanitized name that complies with AWS requirements.

        Raises:
            RuntimeError: If the random suffix is invalid.
            ValueError: If the service name is invalid.
        """
        if (
            not re.match(r"^[a-zA-Z0-9]+$", random_suffix)
            or len(random_suffix) < 4
        ):
            raise RuntimeError(
                f"Invalid random suffix: {random_suffix}. Must contain only "
                "letters and numbers and be at least 4 characters long."
            )

        # Use the first extra allowed character as the separator
        separator = extra_allowed_characters[0]

        # Replace all disallowed characters with the separator
        sanitized = re.sub(
            rf"[^a-zA-Z0-9{extra_allowed_characters}]",
            separator,
            name,
        )

        # Remove consecutive extra allowed characters
        for char in extra_allowed_characters:
            sanitized = re.sub(
                rf"[{char}]+",
                char,
                sanitized,
            )

        # Remove leading and trailing extra allowed characters before truncating
        sanitized = re.sub(
            r"^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$",
            "",
            sanitized,
        )

        # Truncate to fit within max_length character limit including suffix
        max_base_length = (
            max_length - len(random_suffix) - 1
        )  # -1 for the separator
        if len(sanitized) > max_base_length:
            sanitized = sanitized[:max_base_length]

        # Ensure it starts and ends with alphanumeric characters
        sanitized = re.sub(
            r"^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$",
            "",
            sanitized,
        )

        # Ensure we have at least one character after cleanup
        if not sanitized:
            raise ValueError(
                f"Invalid name: {name}. Must contain at least one "
                "alphanumeric character."
            )

        return f"{sanitized}{separator}{random_suffix}"

    def _get_service_name(
        self, deployment_name: str, deployment_id: UUID, prefix: str
    ) -> str:
        """Get the App Runner service name for a deployment.

        Args:
            deployment_name: The deployment name.
            deployment_id: The deployment ID.
            prefix: The prefix to use for the service name.

        Returns:
            The App Runner service name that complies with all naming
            requirements.
        """
        # We use the first 8 characters of the deployment UUID as a random
        # suffix to ensure uniqueness.
        deployment_id_short = str(deployment_id)[:8]
        raw_name = f"{prefix}{deployment_name}"

        return self._sanitize_name(
            raw_name,
            random_suffix=deployment_id_short,
            max_length=40,
            extra_allowed_characters="-_",
        )

    def _get_secret_name(
        self,
        deployment_name: str,
        deployment_id: UUID,
        prefix: str,
    ) -> str:
        """Get the Secrets Manager secret name for a deployment.

        Args:
            deployment_name: The deployment name.
            deployment_id: The deployment ID.
            prefix: The prefix to use for the secret name.

        Returns:
            The Secrets Manager secret name.
        """
        # We use the first 8 characters of the deployment UUID as a random
        # suffix to ensure uniqueness.
        deployment_id_short = str(deployment_id)[:8]
        raw_name = f"{prefix}{deployment_name}"

        return self._sanitize_name(
            raw_name,
            random_suffix=deployment_id_short,
            max_length=512,
            extra_allowed_characters="-_./",
        )

    def _create_or_update_secret(
        self,
        secret_name: str,
        secret_value: str,
        deployment: DeploymentResponse,
        settings: AWSDeployerSettings,
    ) -> str:
        """Create or update a secret in Secrets Manager.

        Args:
            secret_name: The name of the secret.
            secret_value: The value to store.
            deployment: The deployment.
            settings: The deployer settings.

        Returns:
            The secret ARN.

        Raises:
            ClientError: If the secret cannot be updated.
            DeployerError: If secret creation/update fails.
        """
        try:
            try:
                response = self.secrets_manager_client.update_secret(
                    SecretId=secret_name,
                    SecretString=secret_value,
                )
                self.secrets_manager_client.tag_resource(
                    SecretId=response["ARN"],
                    Tags=self.get_tags(deployment, settings),
                )
                logger.debug(f"Updated existing secret {secret_name}")
                return response["ARN"]  # type: ignore[no-any-return]
            except ClientError as e:
                if e.response["Error"]["Code"] == "ResourceNotFoundException":
                    logger.debug(f"Creating new secret {secret_name}")
                    response = self.secrets_manager_client.create_secret(
                        Name=secret_name,
                        SecretString=secret_value,
                        Description=f"ZenML deployment secret for {deployment.name}",
                        Tags=self.get_tags(deployment, settings),
                    )
                    logger.debug(f"Created new secret {secret_name}")
                    return response["ARN"]  # type: ignore[no-any-return]
                else:
                    raise

        except (ClientError, BotoCoreError) as e:
            raise DeployerError(
                f"Failed to create/update secret {secret_name}: {e}"
            )

    def _get_secret_arn(self, deployment: DeploymentResponse) -> Optional[str]:
        """Get the existing AWS Secrets Manager secret ARN for a deployment.

        Args:
            deployment: The deployment.

        Returns:
            The existing AWS Secrets Manager secret ARN for the deployment,
            or None if no secret exists.
        """
        metadata = AppRunnerDeploymentMetadata.from_deployment(deployment)

        if not metadata.secret_arn:
            return None

        try:
            self.secrets_manager_client.describe_secret(
                SecretId=metadata.secret_arn
            )
            return metadata.secret_arn
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return None
            logger.exception(f"Failed to verify secret {metadata.secret_arn}")
            return None

    def _delete_secret(self, secret_arn: str) -> None:
        """Delete a secret from Secrets Manager.

        Args:
            secret_arn: The ARN of the secret to delete.
        """
        try:
            self.secrets_manager_client.delete_secret(
                SecretId=secret_arn,
                ForceDeleteWithoutRecovery=True,
            )
            logger.debug(f"Deleted secret {secret_arn}")
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                logger.debug(
                    f"Secret {secret_arn} not found, skipping deletion"
                )
            else:
                logger.exception(f"Failed to delete secret {secret_arn}")

    def _cleanup_deployment_secrets(
        self,
        deployment: DeploymentResponse,
    ) -> None:
        """Clean up the secret associated with a deployment.

        Args:
            deployment: The deployment.
        """
        secret_arn = self._get_secret_arn(deployment)

        if secret_arn:
            self._delete_secret(secret_arn)

    def _get_auto_scaling_config_name(
        self, deployment_name: str, deployment_id: UUID
    ) -> str:
        """Get the auto-scaling configuration name for a deployment.

        Args:
            deployment_name: The deployment name.
            deployment_id: The deployment ID.

        Returns:
            The auto-scaling configuration name.
        """
        # We use the first 8 characters of the deployment UUID as a random
        # suffix to ensure uniqueness.
        deployment_id_short = str(deployment_id)[:8]
        raw_name = f"zenml-{deployment_name}-{deployment_id_short}"

        return self._sanitize_name(
            raw_name,
            random_suffix=deployment_id_short,
            max_length=32,
            extra_allowed_characters="-_",
        )

    def _create_or_update_auto_scaling_config(
        self,
        config_name: str,
        min_size: int,
        max_size: int,
        max_concurrency: int,
        deployment: DeploymentResponse,
        settings: AWSDeployerSettings,
    ) -> str:
        """Create or update an auto-scaling configuration for App Runner.

        Args:
            config_name: The name for the auto-scaling configuration.
            min_size: Minimum number of instances.
            max_size: Maximum number of instances.
            max_concurrency: Maximum concurrent requests per instance.
            deployment: The deployment.
            settings: The deployer settings.

        Returns:
            The ARN of the created/updated auto-scaling configuration.

        Raises:
            ClientError: If the auto-scaling configuration cannot be described.
            DeployerError: If auto-scaling configuration creation/update fails.
        """
        try:
            metadata = AppRunnerDeploymentMetadata.from_deployment(deployment)
            existing_arn = metadata.auto_scaling_configuration_arn

            if existing_arn:
                try:
                    response = self.app_runner_client.describe_auto_scaling_configuration(
                        AutoScalingConfigurationArn=existing_arn
                    )
                    existing_config = response["AutoScalingConfiguration"]

                    if (
                        existing_config["MaxConcurrency"] == max_concurrency
                        and existing_config["MaxSize"] == max_size
                        and existing_config["MinSize"] == min_size
                    ):
                        logger.debug(
                            f"Auto-scaling configuration {existing_arn} is up "
                            "to date"
                        )
                        return existing_arn

                    logger.debug(
                        f"Auto-scaling configuration {existing_arn} is out of "
                        "date, updating it"
                    )

                except ClientError as e:
                    if (
                        e.response["Error"]["Code"]
                        != "InvalidRequestException"
                    ):
                        raise
                    logger.debug(
                        f"Existing auto-scaling configuration {existing_arn} "
                        "not found, creating new one"
                    )
            else:
                logger.debug(
                    f"Creating auto-scaling configuration {config_name}"
                )

            # The create_auto_scaling_configuration call is used to both create
            # a new auto-scaling configuration and update an existing one.
            # It is possible to create multiple revisions of the same
            # configuration by calling create_auto_scaling_configuration
            # multiple times using the same AutoScalingConfigurationName.
            response = (
                self.app_runner_client.create_auto_scaling_configuration(
                    AutoScalingConfigurationName=config_name,
                    MaxConcurrency=max_concurrency,
                    MaxSize=max_size,
                    MinSize=min_size,
                    Tags=self.get_tags(deployment, settings),
                )
            )

            return response["AutoScalingConfiguration"][  # type: ignore[no-any-return]
                "AutoScalingConfigurationArn"
            ]

        except (ClientError, BotoCoreError) as e:
            raise DeployerError(
                f"Failed to create/update auto-scaling configuration "
                f"{config_name}: {e}"
            )

    def _cleanup_deployment_auto_scaling_config(
        self, deployment: DeploymentResponse
    ) -> None:
        """Clean up the auto-scaling configuration associated with a deployment.

        Args:
            deployment: The deployment.
        """
        metadata = AppRunnerDeploymentMetadata.from_deployment(deployment)
        config_arn = metadata.auto_scaling_configuration_arn
        if not config_arn:
            return

        try:
            logger.debug(f"Deleting auto-scaling configuration {config_arn}")
            self.app_runner_client.delete_auto_scaling_configuration(
                AutoScalingConfigurationArn=config_arn
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                logger.debug(
                    f"Auto-scaling configuration {config_arn} not found, "
                    "skipping deletion"
                )
            else:
                logger.warning(
                    f"Failed to delete auto-scaling configuration {config_arn}: "
                    f"{e}"
                )
        except Exception as e:
            logger.warning(
                f"Failed to delete auto-scaling configuration {config_arn}: "
                f"{e}"
            )

    def _prepare_environment_variables(
        self,
        deployment: DeploymentResponse,
        environment: Dict[str, str],
        secrets: Dict[str, str],
        settings: AWSDeployerSettings,
    ) -> Tuple[Dict[str, str], Dict[str, str], Optional[str]]:
        """Prepare environment variables for App Runner, handling secrets appropriately.

        Args:
            deployment: The deployment.
            environment: Regular environment variables.
            secrets: Sensitive environment variables.
            settings: The deployer settings.

        Returns:
            Tuple containing:
            - Dictionary of regular environment variables.
            - Dictionary of secret environment variables (key -> secret ARN).
            - Optional secret ARN (None if no secrets or fallback to env vars).
        """
        secret_refs = {}
        active_secret_arn: Optional[str] = None

        env_vars = {**settings.environment_variables, **environment}

        if secrets:
            if settings.use_secrets_manager:
                # Always store secrets as single JSON secret and reference their
                # keys in the App Runner service configuration environment
                # variables.

                secret_name = self._get_secret_name(
                    deployment.name, deployment.id, settings.secret_name_prefix
                )

                try:
                    secret_value = json.dumps(secrets)
                    secret_arn = self._create_or_update_secret(
                        secret_name, secret_value, deployment, settings
                    )
                    active_secret_arn = secret_arn

                    for key in secrets.keys():
                        secret_refs[key] = f"{secret_arn}:{key}::"

                    logger.debug(
                        f"Secret {secret_name} stored with ARN {secret_arn} "
                        f"containing {len(secrets)} secret(s)"
                    )

                except Exception as e:
                    logger.warning(
                        f"Failed to create secret, falling back "
                        f"to direct env vars: {e}"
                    )
                    env_vars.update(secrets)

                existing_secret_arn = self._get_secret_arn(deployment)
                if (
                    existing_secret_arn
                    and existing_secret_arn != active_secret_arn
                ):
                    # Sometimes the previous secret resource is different from
                    # the new secret resource, e.g. if the secret name changed.
                    # In this case, we need to delete the old secret resource.
                    self._delete_secret(existing_secret_arn)
            else:
                logger.warning(
                    "Storing secrets directly in environment variables. "
                    "Consider enabling use_secrets_manager for better security."
                )
                env_vars.update(secrets)

        return env_vars, secret_refs, active_secret_arn

    def _get_app_runner_service(
        self, deployment: DeploymentResponse
    ) -> Optional[Dict[str, Any]]:
        """Get an existing App Runner service for a deployment.

        Args:
            deployment: The deployment.

        Returns:
            The App Runner service dictionary, or None if it doesn't exist.

        Raises:
            ClientError: If the App Runner service cannot be described.
        """
        existing_metadata = AppRunnerDeploymentMetadata.from_deployment(
            deployment
        )

        if not existing_metadata.service_arn:
            return None

        try:
            response = self.app_runner_client.describe_service(
                ServiceArn=existing_metadata.service_arn
            )
            return response["Service"]  # type: ignore[no-any-return]
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return None
            raise

    def _get_service_operational_state(
        self,
        service: Dict[str, Any],
        region: str,
        secret_arn: Optional[str] = None,
    ) -> DeploymentOperationalState:
        """Get the operational state of an App Runner service.

        Args:
            service: The App Runner service dictionary.
            region: The AWS region.
            secret_arn: The active Secrets Manager secret ARN.

        Returns:
            The operational state of the App Runner service.
        """
        metadata = AppRunnerDeploymentMetadata.from_app_runner_service(
            service, region, secret_arn
        )

        state = DeploymentOperationalState(
            status=DeploymentStatus.UNKNOWN,
            metadata=metadata.model_dump(exclude_none=True),
        )

        # Map App Runner service status to ZenML status. Valid values are:
        # - CREATE_FAILED
        # - DELETE_FAILED
        # - RUNNING
        # - DELETED
        # - PAUSED
        # - OPERATION_IN_PROGRESS
        service_status = service.get("Status", "").upper()

        if service_status in [
            "CREATE_FAILED",
            "DELETE_FAILED",
        ]:
            state.status = DeploymentStatus.ERROR
        elif service_status == "OPERATION_IN_PROGRESS":
            state.status = DeploymentStatus.PENDING
        elif service_status == "RUNNING":
            state.status = DeploymentStatus.RUNNING
            state.url = service.get("ServiceUrl")
            if state.url and not state.url.startswith("https://"):
                state.url = f"https://{state.url}"
        elif service_status == "DELETED":
            state.status = DeploymentStatus.ABSENT
        elif service_status == "PAUSED":
            state.status = (
                DeploymentStatus.PENDING
            )  # Treat paused as pending for now
        else:
            state.status = DeploymentStatus.UNKNOWN

        return state

    def _requires_service_replacement(
        self,
        existing_service: Dict[str, Any],
        settings: AWSDeployerSettings,
    ) -> bool:
        """Check if the service configuration requires replacement.

        App Runner only requires service replacement for fundamental service-level
        changes that cannot be handled through revisions:

        - Network access configuration
        - VPC configuration
        - Encryption configuration
        - Observability configuration

        All other configuration changes (image, resources, environment, scaling)
        can be handled as updates.

        Args:
            existing_service: The existing App Runner service.
            settings: The new deployer settings.

        Returns:
            True if the service needs to be replaced, False if it can be updated.
        """
        network_config = existing_service.get("NetworkConfiguration", {})
        ingress_config = network_config.get("IngressConfiguration", {})
        current_public_access = ingress_config.get("IsPubliclyAccessible")
        if current_public_access != settings.is_publicly_accessible:
            return True

        current_vpc_config = network_config.get("EgressConfiguration", {})
        has_current_vpc = bool(current_vpc_config.get("VpcConnectorArn"))
        will_have_vpc = bool(settings.ingress_vpc_configuration)
        if has_current_vpc != will_have_vpc:
            return True

        current_encryption = existing_service.get(
            "EncryptionConfiguration", {}
        )
        current_kms_key = current_encryption.get("KmsKey")
        if current_kms_key != settings.encryption_kms_key:
            return True

        return False

    def _convert_resource_settings_to_aws_format(
        self,
        resource_settings: ResourceSettings,
        resource_combinations: List[Tuple[float, float]],
        strict_resource_matching: bool = False,
    ) -> Tuple[str, str]:
        """Convert ResourceSettings to AWS App Runner resource format.

        AWS App Runner only supports specific CPU-memory combinations.
        This method selects the best combination that meets the requirements.

        Args:
            resource_settings: The resource settings from pipeline configuration.
            resource_combinations: List of supported CPU (vCPU) and memory (GB)
                combinations.
            strict_resource_matching: Whether to enforce strict matching of
                resource requirements to AWS App Runner supported CPU and
                memory combinations or approximate the closest matching
                supported combination.

        Returns:
            Tuple of (cpu, memory) in AWS App Runner format.
        """
        requested_cpu = resource_settings.cpu_count
        requested_memory_gb = None
        if resource_settings.memory is not None:
            requested_memory_gb = resource_settings.get_memory(unit="GB")

        cpu, memory = self._select_aws_cpu_memory_combination(
            requested_cpu,
            requested_memory_gb,
            resource_combinations,
            strict_resource_matching,
        )

        return cpu, memory

    def _select_aws_cpu_memory_combination(
        self,
        requested_cpu: Optional[float],
        requested_memory_gb: Optional[float],
        resource_combinations: List[Tuple[float, float]],
        strict_resource_matching: bool = False,
    ) -> Tuple[str, str]:
        """Select the best AWS App Runner CPU-memory combination.

        AWS App Runner only supports specific CPU and memory combinations, e.g.:
        - 0.25 vCPU: 0.5 GB, 1 GB
        - 0.5 vCPU: 1 GB
        - 1 vCPU: 2 GB, 3 GB, 4 GB
        - 2 vCPU: 4 GB, 6 GB
        - 4 vCPU: 8 GB, 10 GB, 12 GB

        This method selects the best combination that meets the requirements.

        Args:
            requested_cpu: Requested CPU count (can be None)
            requested_memory_gb: Requested memory in GB (can be None)
            resource_combinations: List of supported CPU (vCPU) and memory (GB)
                combinations.
            strict_resource_matching: Whether to enforce strict matching of
                resource requirements to AWS App Runner supported CPU and
                memory combinations or approximate the closest matching
                supported combination.

        Returns:
            Tuple of (cpu, memory) that best matches requirements, in AWS App
            Runner format.

        Raises:
            ValueError: If the requested resource requirements cannot be matched
                to any of the supported combinations for the AWS App Runner
                service and strict_resource_matching is True.
        """
        if requested_cpu is None and requested_memory_gb is None:
            return f"{DEFAULT_CPU:g} vCPU", f"{DEFAULT_MEMORY:g} GB"

        sorted_combinations = sorted(resource_combinations)

        best_combination = None
        exact_match = False
        best_score = float("inf")  # Lower is better

        for cpu_val, mem_val in sorted_combinations:
            cpu_ok = requested_cpu is None or cpu_val >= requested_cpu
            mem_ok = (
                requested_memory_gb is None or mem_val >= requested_memory_gb
            )
            exact_match = (
                cpu_val == requested_cpu and mem_val == requested_memory_gb
            )
            if exact_match:
                best_combination = (cpu_val, mem_val)
                break

            if cpu_ok and mem_ok:
                # Calculate "waste" score (how much over-provisioning)
                cpu_waste = (
                    0 if requested_cpu is None else (cpu_val - requested_cpu)
                )
                mem_waste = (
                    0
                    if requested_memory_gb is None
                    else (mem_val - requested_memory_gb)
                )

                # Prioritize CPU requirements, then memory
                score = cpu_waste * 10 + mem_waste

                if score < best_score:
                    best_score = score
                    best_combination = (cpu_val, mem_val)

        # If no combination satisfies requirements, use the highest available
        if best_combination is None:
            best_combination = sorted_combinations[-1]

        result = (
            f"{best_combination[0]:g} vCPU",
            f"{best_combination[1]:g} GB",
        )

        if strict_resource_matching and not exact_match:
            raise ValueError(
                f"Requested resource requirements ({requested_cpu} vCPU, "
                f"{requested_memory_gb} GB) cannot be matched to any of the "
                f"supported combinations for the AWS App Runner service. "
                f"The closest matching combination is {result[0]} and "
                f"{result[1]}."
            )

        return result

    def _convert_scaling_settings_to_aws_format(
        self,
        resource_settings: ResourceSettings,
    ) -> Tuple[int, int, int]:
        """Convert ResourceSettings scaling to AWS App Runner format.

        Args:
            resource_settings: The resource settings from pipeline configuration.

        Returns:
            Tuple of (min_replicas, max_replicas, max_concurrency) for AWS App
            Runner.
        """
        min_replicas = DEFAULT_MIN_REPLICAS
        if resource_settings.min_replicas is not None:
            min_replicas = max(
                1, resource_settings.min_replicas
            )  # AWS App Runner min is 1

        max_replicas = DEFAULT_MAX_REPLICAS
        if resource_settings.max_replicas is not None:
            # ResourceSettings uses 0 to mean "no limit"
            # AWS App Runner needs a specific value, so we use the platform maximum
            if resource_settings.max_replicas == 0:
                max_replicas = AWS_APP_RUNNER_MAX_SIZE
            else:
                max_replicas = min(
                    resource_settings.max_replicas, AWS_APP_RUNNER_MAX_SIZE
                )

        max_concurrency = DEFAULT_MAX_CONCURRENCY
        if resource_settings.max_concurrency is not None:
            max_concurrency = min(
                resource_settings.max_concurrency,
                AWS_APP_RUNNER_MAX_CONCURRENCY,
            )

        return min_replicas, max_replicas, max_concurrency

    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Serve a pipeline as an App Runner service.

        Args:
            deployment: The deployment to serve.
            stack: The stack the pipeline will be deployed on.
            environment: Environment variables to set.
            secrets: Secret environment variables to set.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be provisioned.

        Returns:
            The operational state of the provisioned deployment.

        Raises:
            DeploymentProvisionError: If the deployment fails.
            DeploymentDeprovisionError: If the previous deployment fails to
                deprovision.
            DeployerError: If an unexpected error occurs.
        """
        snapshot = deployment.snapshot
        assert snapshot, "Pipeline snapshot not found"

        settings = cast(
            AWSDeployerSettings,
            self.get_settings(snapshot),
        )

        resource_settings = snapshot.pipeline_configuration.resource_settings

        cpu, memory = self._convert_resource_settings_to_aws_format(
            resource_settings,
            self.config.resource_combinations,
            settings.strict_resource_matching,
        )
        min_size, max_size, max_concurrency = (
            self._convert_scaling_settings_to_aws_format(
                resource_settings,
            )
        )

        client = self.app_runner_client

        service_name = self._get_service_name(
            deployment.name, deployment.id, settings.service_name_prefix
        )

        existing_service = self._get_app_runner_service(deployment)
        image = self.get_image(snapshot)
        region = self.region

        if existing_service and self._requires_service_replacement(
            existing_service, settings
        ):
            try:
                self.do_deprovision_deployment(deployment, timeout)
            except DeploymentNotFoundError:
                logger.warning(
                    f"Deployment '{deployment.name}' not found, "
                    f"skipping deprovision of existing App Runner service"
                )
            except DeployerError as e:
                raise DeploymentDeprovisionError(
                    f"Failed to deprovision existing App Runner service for "
                    f"deployment '{deployment.name}': {e}\n"
                    "Bailing out to avoid leaving orphaned resources."
                    "You might need to manually delete the existing App Runner "
                    "service instance to continue or forcefully delete the "
                    "deployment."
                )
            existing_service = None

        entrypoint = DeploymentEntrypointConfiguration.get_entrypoint_command()
        arguments = DeploymentEntrypointConfiguration.get_entrypoint_arguments(
            **{
                DEPLOYMENT_ID_OPTION: deployment.id,
            }
        )

        env_vars, secret_refs, active_secret_arn = (
            self._prepare_environment_variables(
                deployment, environment, secrets, settings
            )
        )

        # AWS App Runner only supports ECR repositories.
        if "public.ecr.aws" in image:
            image_repo_type = "ECR_PUBLIC"
        elif "amazonaws.com" in image:
            image_repo_type = "ECR"
        else:
            raise DeploymentProvisionError(
                f"AWS App Runner only supports Amazon ECR and ECR Public "
                f"repositories. The container image '{image}' does not appear "
                f"to be hosted on either platform. Supported image repositories:\n"
                f"- ECR Public: public.ecr.aws/...\n"
                f"- ECR Private: *.amazonaws.com/...\n"
                f"Please push your image to one of these registries before "
                f"deploying to App Runner."
            )

        container_port = (
            snapshot.pipeline_configuration.deployment_settings.uvicorn_port
        )
        image_config: Dict[str, Any] = {
            "Port": str(container_port),
            "StartCommand": " ".join(entrypoint + arguments),
        }

        if env_vars:
            image_config["RuntimeEnvironmentVariables"] = env_vars

        if secret_refs:
            image_config["RuntimeEnvironmentSecrets"] = secret_refs

        image_repository_config = {
            "ImageIdentifier": image,
            "ImageConfiguration": image_config,
            "ImageRepositoryType": image_repo_type,
        }

        source_configuration = {
            "ImageRepository": image_repository_config,
            # We don't want to automatically deploy new revisions when new
            # container images are pushed to the repository.
            "AutoDeploymentsEnabled": False,
        }

        if settings.access_role_arn:
            source_configuration["AuthenticationConfiguration"] = {
                "AccessRoleArn": settings.access_role_arn
            }
        elif image_repo_type == "ECR":
            logger.warning(
                "Using private ECR repository without explicit access_role_arn. "
                "Ensure the default App Runner service role has permissions to "
                f"pull the '{image}' image from the repository, or specify "
                "access_role_arn in deployer settings."
            )

        instance_configuration = {
            "Cpu": cpu,
            "Memory": memory,
        }
        if settings.instance_role_arn:
            instance_configuration["InstanceRoleArn"] = (
                settings.instance_role_arn
            )
        elif secret_refs:
            logger.warning(
                "Storing secrets in AWS Secrets Manager is enabled but no "
                "explicit instance role is provided. Ensure the default "
                "App Runner service role has secretsmanager:GetSecretValue "
                "permissions, provide an explicit instance role or disable "
                "'use_secrets_manager' in deployer settings."
            )

        auto_scaling_config_name = self._get_auto_scaling_config_name(
            deployment.name, deployment.id
        )
        auto_scaling_config_arn = self._create_or_update_auto_scaling_config(
            auto_scaling_config_name,
            min_size,
            max_size,
            max_concurrency,
            deployment,
            settings,
        )

        health_check_configuration = {
            "Protocol": settings.health_check_protocol,
            "Interval": settings.health_check_interval_seconds,
            "Timeout": settings.health_check_timeout_seconds,
            "HealthyThreshold": settings.health_check_healthy_threshold,
            "UnhealthyThreshold": settings.health_check_unhealthy_threshold,
        }

        deployment_settings = (
            snapshot.pipeline_configuration.deployment_settings
        )

        if settings.health_check_protocol.upper() == "HTTP":
            root_path = deployment_settings.root_url_path
            api_url_path = deployment_settings.api_url_path
            health_check_path = f"{root_path}{api_url_path}{deployment_settings.health_url_path}"
            health_check_configuration["Path"] = health_check_path

        network_configuration = {
            "IngressConfiguration": {
                "IsPubliclyAccessible": settings.is_publicly_accessible,
            }
        }

        traffic_configurations = []
        for revision, percent in settings.traffic_allocation.items():
            if revision == "LATEST":
                traffic_configurations.append(
                    {
                        "Type": "LATEST",
                        "Percent": percent,
                    }
                )
            else:
                if revision.startswith("tag:"):
                    traffic_configurations.append(
                        {
                            "Tag": revision[4:],  # Remove "tag:" prefix
                            "Percent": percent,
                        }
                    )
                else:
                    traffic_configurations.append(
                        {
                            "Revision": revision,
                            "Percent": percent,
                        }
                    )

        if settings.ingress_vpc_configuration:
            vpc_config = json.loads(settings.ingress_vpc_configuration)
            network_configuration["IngressConfiguration"][
                "VpcIngressConnectionConfiguration"
            ] = vpc_config

        encryption_configuration = None
        if settings.encryption_kms_key:
            encryption_configuration = {
                "KmsKey": settings.encryption_kms_key,
            }

        observability_configuration = None
        if settings.observability_configuration_arn:
            observability_configuration = {
                "ObservabilityEnabled": True,
                "ObservabilityConfigurationArn": settings.observability_configuration_arn,
            }

        service_tags = self.get_tags(deployment, settings)

        try:
            if existing_service:
                logger.debug(
                    f"Updating existing App Runner service for pipeline "
                    f"deployment '{deployment.name}'"
                )

                update_request = {
                    "ServiceArn": existing_service["ServiceArn"],
                    "SourceConfiguration": source_configuration,
                    "InstanceConfiguration": instance_configuration,
                    "AutoScalingConfigurationArn": auto_scaling_config_arn,
                    "HealthCheckConfiguration": health_check_configuration,
                    "NetworkConfiguration": network_configuration,
                }

                if not (
                    len(traffic_configurations) == 1
                    and traffic_configurations[0].get("Type") == "LATEST"
                    and traffic_configurations[0].get("Percent") == 100
                ):
                    update_request["TrafficConfiguration"] = (
                        traffic_configurations
                    )

                if encryption_configuration:
                    update_request["EncryptionConfiguration"] = (
                        encryption_configuration
                    )

                if observability_configuration:
                    update_request["ObservabilityConfiguration"] = (
                        observability_configuration
                    )

                response = client.update_service(**update_request)
                service_arn = response["Service"]["ServiceArn"]

                # Update tags separately
                client.tag_resource(
                    ResourceArn=service_arn,
                    Tags=service_tags,
                )

                updated_service = response["Service"]
            else:
                logger.debug(
                    f"Creating new App Runner service for deployment "
                    f"'{deployment.name}' in region {region}"
                )

                create_request = {
                    "ServiceName": service_name,
                    "SourceConfiguration": source_configuration,
                    "InstanceConfiguration": instance_configuration,
                    "AutoScalingConfigurationArn": auto_scaling_config_arn,
                    "Tags": service_tags,
                    "HealthCheckConfiguration": health_check_configuration,
                    "NetworkConfiguration": network_configuration,
                }

                if encryption_configuration:
                    create_request["EncryptionConfiguration"] = (
                        encryption_configuration
                    )

                if observability_configuration:
                    create_request["ObservabilityConfiguration"] = (
                        observability_configuration
                    )

                # Only add traffic configuration if it's not the default
                # (100% LATEST)
                if not (
                    len(traffic_configurations) == 1
                    and traffic_configurations[0].get("Type") == "LATEST"
                    and traffic_configurations[0].get("Percent") == 100
                ):
                    create_request["TrafficConfiguration"] = (
                        traffic_configurations
                    )

                response = client.create_service(**create_request)
                updated_service = response["Service"]

            return self._get_service_operational_state(
                updated_service, region, active_secret_arn
            )

        except (ClientError, BotoCoreError) as e:
            raise DeploymentProvisionError(
                f"Failed to deploy App Runner service for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while provisioning deployment "
                f"'{deployment.name}': {e}"
            )

    def do_get_deployment_state(
        self,
        deployment: DeploymentResponse,
    ) -> DeploymentOperationalState:
        """Get information about an App Runner deployment.

        Args:
            deployment: The deployment to get information about.

        Returns:
            The operational state of the deployment.

        Raises:
            DeploymentNotFoundError: If the deployment is not found.
            RuntimeError: If the service ARN is not found in the deployment metadata.
        """
        service = self._get_app_runner_service(deployment)

        if service is None:
            raise DeploymentNotFoundError(
                f"App Runner service for deployment '{deployment.name}' "
                "not found"
            )

        existing_metadata = AppRunnerDeploymentMetadata.from_deployment(
            deployment
        )

        if not existing_metadata.region:
            raise RuntimeError(
                f"Region not found in deployment metadata for "
                f"deployment '{deployment.name}'"
            )

        existing_secret_arn = self._get_secret_arn(deployment)

        return self._get_service_operational_state(
            service,
            existing_metadata.region,
            existing_secret_arn,
        )

    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Get the logs of an App Runner deployment.

        Args:
            deployment: The deployment to get the logs of.
            follow: If True, stream logs as they are written.
            tail: Only retrieve the last NUM lines of log output.

        Yields:
            The logs of the deployment.

        Raises:
            NotImplementedError: If log following is requested.
            DeploymentNotFoundError: If the deployment is not found.
            DeploymentLogsNotFoundError: If the logs are not found.
            DeployerError: If an unexpected error occurs.
            RuntimeError: If the service name is not found in the deployment metadata.
        """
        if follow:
            raise NotImplementedError(
                "Log following is not yet implemented for App Runner deployer"
            )

        service = self._get_app_runner_service(deployment)
        if service is None:
            raise DeploymentNotFoundError(
                f"App Runner service for deployment '{deployment.name}' not "
                "found"
            )

        try:
            existing_metadata = AppRunnerDeploymentMetadata.from_deployment(
                deployment
            )
            service_name = existing_metadata.service_name
            service_id = existing_metadata.service_id
            if not service_name or not service_id:
                raise RuntimeError(
                    f"Service name or ID not found in deployment metadata for "
                    f"deployment '{deployment.name}'"
                )

            # App Runner automatically creates CloudWatch log groups
            log_group_name = (
                f"/aws/apprunner/{service_name}/{service_id}/application"
            )

            try:
                streams_response = self.logs_client.describe_log_streams(
                    logGroupName=log_group_name,
                    orderBy="LastEventTime",
                    descending=True,
                )

                log_lines = []
                for stream in streams_response.get("logStreams", []):
                    stream_name = stream["logStreamName"]

                    events_response = self.logs_client.get_log_events(
                        logGroupName=log_group_name,
                        logStreamName=stream_name,
                        startFromHead=False,  # Get most recent first
                    )

                    for event in events_response.get("events", []):
                        timestamp = event.get("timestamp", 0)
                        message = event.get("message", "")

                        dt = datetime.datetime.fromtimestamp(
                            timestamp / 1000.0
                        )
                        formatted_time = dt.isoformat()

                        log_line = f"[{formatted_time}] {message}"
                        log_lines.append(log_line)

                # Sort by timestamp (most recent last for tail to work correctly)
                log_lines.sort()

                if tail is not None and tail > 0:
                    log_lines = log_lines[-tail:]

                for log_line in log_lines:
                    yield log_line

            except ClientError as e:
                if e.response["Error"]["Code"] == "ResourceNotFoundException":
                    raise DeploymentLogsNotFoundError(
                        f"Log group not found for App Runner service "
                        f"'{service_name}'"
                    )
                raise

        except (ClientError, BotoCoreError) as e:
            raise DeploymentLogsNotFoundError(
                f"Failed to retrieve logs for deployment '{deployment.name}': "
                f"{e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while retrieving logs for deployment "
                f"'{deployment.name}': {e}"
            )

    def do_deprovision_deployment(
        self,
        deployment: DeploymentResponse,
        timeout: int,
    ) -> Optional[DeploymentOperationalState]:
        """Deprovision an App Runner deployment.

        Args:
            deployment: The deployment to deprovision.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be deprovisioned.

        Returns:
            The operational state of the deprovisioned deployment, or None if
            deletion is completed immediately.

        Raises:
            DeploymentNotFoundError: If the deployment is not found.
            DeploymentDeprovisionError: If the deprovision fails.
            DeployerError: If an unexpected error occurs.
            RuntimeError: If the service ARN is not found in the deployment metadata.
        """
        service = self._get_app_runner_service(deployment)
        if service is None:
            raise DeploymentNotFoundError(
                f"App Runner service for deployment '{deployment.name}' not "
                "found"
            )

        try:
            existing_metadata = AppRunnerDeploymentMetadata.from_deployment(
                deployment
            )
            if not existing_metadata.service_arn:
                raise RuntimeError(
                    f"Service ARN not found in deployment metadata for "
                    f"deployment '{deployment.name}'"
                )

            logger.debug(
                f"Deleting App Runner service for deployment "
                f"'{deployment.name}'"
            )

            # Delete the service
            self.app_runner_client.delete_service(
                ServiceArn=existing_metadata.service_arn
            )

        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                raise DeploymentNotFoundError(
                    f"App Runner service for deployment '{deployment.name}' "
                    "not found"
                )
            raise DeploymentDeprovisionError(
                f"Failed to delete App Runner service for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while deleting deployment "
                f"'{deployment.name}': {e}"
            )

        deployment_before_deletion = deployment

        # App Runner deletion is asynchronous and the auto-scaling configuration
        # and secrets need to be cleaned up after the service is deleted. So we
        # poll the service here instead of doing it in the base deployer class.
        deployment, deployment_state = self._poll_deployment(
            deployment, DeploymentStatus.ABSENT, timeout
        )

        if deployment_state.status != DeploymentStatus.ABSENT:
            return deployment_state

        try:
            self._cleanup_deployment_secrets(deployment_before_deletion)

            self._cleanup_deployment_auto_scaling_config(
                deployment_before_deletion
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while cleaning up resources for pipeline "
                f"deployment '{deployment.name}': {e}"
            )

        return None



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of artifact table."""

from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple
from uuid import UUID

from pydantic import ValidationError
from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlalchemy.orm import joinedload, object_session
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship, asc, col, desc, select

from zenml.config.source import Source
from zenml.enums import (
    ArtifactSaveType,
    ArtifactType,
    ExecutionStatus,
    MetadataResourceTypes,
    TaggableResourceTypes,
)
from zenml.models import (
    ArtifactResponse,
    ArtifactResponseBody,
    ArtifactResponseMetadata,
    ArtifactResponseResources,
    ArtifactUpdate,
    ArtifactVersionRequest,
    ArtifactVersionResponse,
    ArtifactVersionResponseBody,
    ArtifactVersionResponseMetadata,
    ArtifactVersionResponseResources,
    ArtifactVersionUpdate,
)
from zenml.models.v2.core.artifact import ArtifactRequest
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import BaseSchema, NamedSchema
from zenml.zen_stores.schemas.component_schemas import StackComponentSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import (
    RunMetadataInterface,
    jl_arg,
)

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.artifact_visualization_schemas import (
        ArtifactVisualizationSchema,
    )
    from zenml.zen_stores.schemas.model_schemas import (
        ModelVersionArtifactSchema,
    )
    from zenml.zen_stores.schemas.run_metadata_schemas import RunMetadataSchema
    from zenml.zen_stores.schemas.step_run_schemas import (
        StepRunInputArtifactSchema,
        StepRunOutputArtifactSchema,
    )
    from zenml.zen_stores.schemas.tag_schemas import TagSchema


class ArtifactSchema(NamedSchema, table=True):
    """SQL Model for artifacts."""

    __tablename__ = "artifact"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_artifact_name_in_project",
        ),
    )

    # Fields
    has_custom_name: bool
    versions: List["ArtifactVersionSchema"] = Relationship(
        back_populates="artifact",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    tags: List["TagSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=f"and_(foreign(TagResourceSchema.resource_type)=='{TaggableResourceTypes.ARTIFACT.value}', foreign(TagResourceSchema.resource_id)==ArtifactSchema.id)",
            secondary="tag_resource",
            secondaryjoin="TagSchema.id == foreign(TagResourceSchema.tag_id)",
            order_by="TagSchema.name",
            overlaps="tags",
        ),
    )

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    project: "ProjectSchema" = Relationship()

    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    user: Optional["UserSchema"] = Relationship(
        back_populates="artifacts",
    )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(ArtifactSchema.user)),
                    # joinedload(jl_arg(ArtifactSchema.tags)),
                ]
            )

        return options

    @property
    def latest_version(self) -> Optional["ArtifactVersionSchema"]:
        """Fetch the latest version for this artifact.

        Raises:
            RuntimeError: If no session for the schema exists.

        Returns:
            The latest version for this artifact.
        """
        if session := object_session(self):
            return (
                session.execute(
                    select(ArtifactVersionSchema)
                    .where(ArtifactVersionSchema.artifact_id == self.id)
                    .order_by(desc(ArtifactVersionSchema.created))
                    .limit(1)
                )
                .scalars()
                .one_or_none()
            )
        else:
            raise RuntimeError(
                "Missing DB session to fetch latest version for artifact."
            )

    @classmethod
    def from_request(
        cls,
        artifact_request: ArtifactRequest,
    ) -> "ArtifactSchema":
        """Convert an `ArtifactRequest` to an `ArtifactSchema`.

        Args:
            artifact_request: The request model to convert.

        Returns:
            The converted schema.
        """
        return cls(
            name=artifact_request.name,
            has_custom_name=artifact_request.has_custom_name,
            project_id=artifact_request.project,
            user_id=artifact_request.user,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> ArtifactResponse:
        """Convert an `ArtifactSchema` to an `ArtifactResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic



        Returns:
            The created `ArtifactResponse`.
        """
        # Create the body of the model
        body = ArtifactResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            created=self.created,
            updated=self.updated,
        )

        # Create the metadata of the model
        metadata = None
        if include_metadata:
            metadata = ArtifactResponseMetadata(
                has_custom_name=self.has_custom_name,
            )

        resources = None
        if include_resources:
            latest_id, latest_name = None, None
            if latest_version := self.latest_version:
                latest_id = latest_version.id
                latest_name = latest_version.version

            resources = ArtifactResponseResources(
                user=self.user.to_model() if self.user else None,
                tags=[tag.to_model() for tag in self.tags],
                latest_version_id=latest_id,
                latest_version_name=latest_name,
            )

        return ArtifactResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(self, artifact_update: ArtifactUpdate) -> "ArtifactSchema":
        """Update an `ArtifactSchema` with an `ArtifactUpdate`.

        Args:
            artifact_update: The update model to apply.

        Returns:
            The updated `ArtifactSchema`.
        """
        self.updated = utc_now()
        if artifact_update.name:
            self.name = artifact_update.name
            self.has_custom_name = True
        if artifact_update.has_custom_name is not None:
            self.has_custom_name = artifact_update.has_custom_name
        return self


class ArtifactVersionSchema(BaseSchema, RunMetadataInterface, table=True):
    """SQL Model for artifact versions."""

    __tablename__ = "artifact_version"
    __table_args__ = (
        UniqueConstraint(
            "version",
            "artifact_id",
            name="unique_version_for_artifact_id",
        ),
    )

    # Fields
    version: str
    version_number: Optional[int]
    type: str
    uri: str = Field(sa_column=Column(TEXT, nullable=False))
    materializer: str = Field(sa_column=Column(TEXT, nullable=False))
    data_type: str = Field(sa_column=Column(TEXT, nullable=False))
    tags: List["TagSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=f"and_(foreign(TagResourceSchema.resource_type)=='{TaggableResourceTypes.ARTIFACT_VERSION.value}', foreign(TagResourceSchema.resource_id)==ArtifactVersionSchema.id)",
            secondary="tag_resource",
            secondaryjoin="TagSchema.id == foreign(TagResourceSchema.tag_id)",
            order_by="TagSchema.name",
            overlaps="tags",
        ),
    )
    save_type: str = Field(sa_column=Column(TEXT, nullable=False))
    content_hash: Optional[str] = Field(sa_column=Column(TEXT, nullable=True))
    item_count: Optional[int] = None

    # Foreign keys
    artifact_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ArtifactSchema.__tablename__,
        source_column="artifact_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    artifact_store_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=StackComponentSchema.__tablename__,
        source_column="artifact_store_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )

    # Relationships
    artifact: "ArtifactSchema" = Relationship(back_populates="versions")
    user: Optional["UserSchema"] = Relationship(
        back_populates="artifact_versions"
    )
    project: "ProjectSchema" = Relationship(back_populates="artifact_versions")
    run_metadata: List["RunMetadataSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            secondary="run_metadata_resource",
            primaryjoin=f"and_(foreign(RunMetadataResourceSchema.resource_type)=='{MetadataResourceTypes.ARTIFACT_VERSION.value}', foreign(RunMetadataResourceSchema.resource_id)==ArtifactVersionSchema.id)",
            secondaryjoin="RunMetadataSchema.id==foreign(RunMetadataResourceSchema.run_metadata_id)",
            overlaps="run_metadata",
        ),
    )
    visualizations: List["ArtifactVisualizationSchema"] = Relationship(
        back_populates="artifact_version",
        sa_relationship_kwargs={"cascade": "delete"},
    )

    # Needed for cascade deletion behavior
    model_versions_artifacts_links: List["ModelVersionArtifactSchema"] = (
        Relationship(
            back_populates="artifact_version",
            sa_relationship_kwargs={"cascade": "delete"},
        )
    )
    output_of_step_runs: List["StepRunOutputArtifactSchema"] = Relationship(
        back_populates="artifact_version",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    input_of_step_runs: List["StepRunInputArtifactSchema"] = Relationship(
        back_populates="artifact_version",
        sa_relationship_kwargs={"cascade": "delete"},
    )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        # if include_metadata:
        #     options.extend(
        #         [
        #             joinedload(jl_arg(ArtifactVersionSchema.visualizations)),
        #             joinedload(jl_arg(ArtifactVersionSchema.run_metadata)),
        #         ]
        #     )

        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(ArtifactVersionSchema.user)),
                    # joinedload(jl_arg(ArtifactVersionSchema.tags)),
                ]
            )

        return options

    @property
    def producer_run_ids(self) -> Optional[Tuple[UUID, UUID]]:
        """Fetch the producer run IDs for this artifact version.

        Raises:
            RuntimeError: If no session for the schema exists.

        Returns:
            The producer step run ID and pipeline run ID for this artifact
            version.
        """
        from zenml.zen_stores.schemas import (
            StepRunOutputArtifactSchema,
            StepRunSchema,
        )

        if session := object_session(self):
            row = session.execute(
                select(StepRunSchema.id, StepRunSchema.pipeline_run_id)
                .join(
                    StepRunOutputArtifactSchema,
                    col(StepRunOutputArtifactSchema.step_id)
                    == col(StepRunSchema.id),
                )
                .where(col(StepRunOutputArtifactSchema.artifact_id) == self.id)
                .where(
                    col(StepRunSchema.status)
                    == ExecutionStatus.COMPLETED.value
                )
                # Fetch the oldest step run
                .order_by(asc(StepRunSchema.created))
                .limit(1)
            ).one_or_none()

            return (row[0], row[1]) if row else None
        else:
            raise RuntimeError(
                "Missing DB session to fetch producer run for artifact version."
            )

    @classmethod
    def from_request(
        cls,
        artifact_version_request: ArtifactVersionRequest,
    ) -> "ArtifactVersionSchema":
        """Convert an `ArtifactVersionRequest` to an `ArtifactVersionSchema`.

        Args:
            artifact_version_request: The request model to convert.

        Raises:
            ValueError: If the request does not specify a version number.

        Returns:
            The converted schema.
        """
        if not artifact_version_request.version:
            raise ValueError("Missing version for artifact version request.")

        try:
            version_number = int(artifact_version_request.version)
        except ValueError:
            version_number = None
        return cls(
            artifact_id=artifact_version_request.artifact_id,
            version=str(artifact_version_request.version),
            version_number=version_number,
            artifact_store_id=artifact_version_request.artifact_store_id,
            project_id=artifact_version_request.project,
            user_id=artifact_version_request.user,
            type=artifact_version_request.type.value,
            uri=artifact_version_request.uri,
            materializer=artifact_version_request.materializer.model_dump_json(),
            data_type=artifact_version_request.data_type.model_dump_json(),
            save_type=artifact_version_request.save_type.value,
            content_hash=artifact_version_request.content_hash,
            item_count=artifact_version_request.item_count,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> ArtifactVersionResponse:
        """Convert an `ArtifactVersionSchema` to an `ArtifactVersionResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic



        Returns:
            The created `ArtifactVersionResponse`.
        """
        try:
            materializer = Source.model_validate_json(self.materializer)
        except ValidationError:
            # This is an old source which was an importable source path
            materializer = Source.from_import_path(self.materializer)

        try:
            data_type = Source.model_validate_json(self.data_type)
        except ValidationError:
            # This is an old source which was an importable source path
            data_type = Source.from_import_path(self.data_type)

        # Create the body of the model
        artifact = self.artifact.to_model()
        body = ArtifactVersionResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            artifact=artifact,
            version=self.version or str(self.version_number),
            uri=self.uri,
            type=ArtifactType(self.type),
            materializer=materializer,
            data_type=data_type,
            created=self.created,
            updated=self.updated,
            save_type=ArtifactSaveType(self.save_type),
            artifact_store_id=self.artifact_store_id,
            content_hash=self.content_hash,
            item_count=self.item_count,
        )

        # Create the metadata of the model
        metadata = None
        if include_metadata:
            metadata = ArtifactVersionResponseMetadata(
                visualizations=[v.to_model() for v in self.visualizations],
                run_metadata=self.fetch_metadata(),
            )

        resources = None
        if include_resources:
            producer_step_run_id, producer_pipeline_run_id = None, None
            if producer_run_ids := self.producer_run_ids:
                # TODO: Why was the producer_pipeline_run_id only set for one
                # of the cases before?
                producer_step_run_id, producer_pipeline_run_id = (
                    producer_run_ids
                )

            resources = ArtifactVersionResponseResources(
                user=self.user.to_model() if self.user else None,
                tags=[tag.to_model() for tag in self.tags],
                producer_step_run_id=producer_step_run_id,
                producer_pipeline_run_id=producer_pipeline_run_id,
            )

        return ArtifactVersionResponse(
            id=self.id,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(
        self, artifact_version_update: ArtifactVersionUpdate
    ) -> "ArtifactVersionSchema":
        """Update an `ArtifactVersionSchema` with an `ArtifactVersionUpdate`.

        Args:
            artifact_version_update: The update model to apply.

        Returns:
            The updated `ArtifactVersionSchema`.
        """
        self.updated = utc_now()
        return self
